#!/usr/bin/env python

"""
Thor HTTP Client

This library allow implementation of an HTTP/1.1 client that is
"non-blocking," "asynchronous" and "event-driven" -- i.e., it achieves very
high performance and concurrency, so long as the application code does not
block (e.g., upon network, disk or database access). Blocking on one response
will block the entire client.

"""

from collections import defaultdict
from urllib.parse import urlsplit, urlunsplit
from typing import Callable, List, Tuple, Union # pylint: disable=unused-import

import thor
from thor.events import EventEmitter, on
from thor.loop import LoopBase, ScheduledEvent
from thor.tcp import TcpClient, TcpConnection
from thor.tls import TlsClient

from thor.http.common import HttpMessageHandler, \
    States, Delimiters, \
    idempotent_methods, no_body_status, hop_by_hop_hdrs, \
    header_names, RawHeaderListType, OriginType
from thor.http.error import UrlError, ConnectError, \
    ReadTimeoutError, HttpVersionError, StartLineError

req_rm_hdrs = hop_by_hop_hdrs + [b'host']

# TODO: next-hop version cache for Expect/Continue, etc.

class HttpClient(object):
    "An asynchronous HTTP client."

    tcp_client_class = TcpClient
    tls_client_class = TlsClient

    def __init__(self, loop: LoopBase=None) -> None:
        self.loop = loop or thor.loop._loop
        self.idle_timeout = 60       # type: int    # seconds
        self.connect_timeout = None  # type: int    # seconds
        self.read_timeout = None     # type: int    # seconds
        self.retry_limit = 2         # type: int
        self.retry_delay = 0.5       # type: float  # seconds
        self.max_server_conn = 4     # type: int
        self.proxy_tls = False       # type: bool
        self.proxy_host = None       # type: bytes
        self.proxy_port = None       # type: int
        self.careful = True          # type: bool
        self._idle_conns = defaultdict(list)     # type: dict[OriginType, list[TcpConnection]]
        self._conn_counts = defaultdict(int)     # type: dict[OriginType, int]
        self.loop.on('stop', self._close_conns)

    def exchange(self) -> 'HttpClientExchange':
        return HttpClientExchange(self)

    def attach_conn(self, origin: OriginType, handle_connect: Callable,
                    handle_connect_error: Callable, connect_timeout: float) -> None:
        "Find an idle connection for origin, or create a new one."
        if self.proxy_host and self.proxy_port:
            # TODO: full form of request-target
            host, port = self.proxy_host, self.proxy_port
            if self.proxy_tls:
                scheme = b'https'
            else:
                scheme = b'http'
            origin = (scheme, host, port)
        else:
            scheme, host, port = origin
        while True:
            try:
                tcp_conn = self._idle_conns[origin].pop()
            except IndexError:
                self._new_conn(origin, handle_connect, handle_connect_error, connect_timeout)
                break
            if tcp_conn.tcp_connected:
                tcp_conn.removeListeners('data', 'pause', 'close')
                tcp_conn.pause(True)
                if hasattr(tcp_conn, "_idler"):
                    tcp_conn._idler.delete() # type: ignore
                handle_connect(tcp_conn)
                break

    def release_conn(self, tcp_conn: TcpConnection, scheme: bytes) -> None:
        "Add an idle connection back to the pool."
        tcp_conn.removeListeners('close')
        origin = (scheme, tcp_conn.host, tcp_conn.port)
        if tcp_conn.tcp_connected:
            def idle_close():
                "Remove the connection from the pool when it closes."
                if hasattr(tcp_conn, "_idler"):
                    tcp_conn._idler.delete()
                self._dead_conn(origin)
                try:
                    self._idle_conns[origin].remove(tcp_conn)
                except (KeyError, ValueError):
                    pass
            tcp_conn.on('close', idle_close)
            if self.idle_timeout > 0:
                tcp_conn._idler = self.loop.schedule(self.idle_timeout, # type: ignore
                                                     tcp_conn.close)
            else:
                tcp_conn.close()
                self._dead_conn(origin)
            self._idle_conns[origin].append(tcp_conn)
        else:
            self._dead_conn(origin)

    def _new_conn(self, origin: OriginType, handle_connect: Callable,
                  handle_error: Callable, timeout: float) -> None:
        "Create a new connection."
        (scheme, host, port) = origin
        tcp_client = None  # type: Union[TcpClient, TlsClient]
        if scheme == b'http':
            tcp_client = self.tcp_client_class(self.loop)
        elif scheme == b'https':
            tcp_client = self.tls_client_class(self.loop)
        else:
            raise ValueError(u'unknown scheme %s' % scheme.decode('utf-8', 'replace'))
        tcp_client.on('connect', handle_connect)
        tcp_client.on('connect_error', handle_error)
        self._conn_counts[origin] += 1
        tcp_client.connect(host, port, timeout)

    def _dead_conn(self, origin: OriginType) -> None:
        "Notify the client that a connect to origin is dead."
        self._conn_counts[origin] -= 1

    def _close_conns(self) -> None:
        "Close all idle HTTP connections."
        for conn_list in self._idle_conns.values():
            for conn in conn_list:
                try:
                    conn.close()
                except:
                    pass
        self._idle_conns.clear()
        # TODO: probably need to close in-progress conns too.


class HttpClientExchange(HttpMessageHandler, EventEmitter):
    default_state = States.QUIET

    def __init__(self, client) -> None:
        HttpMessageHandler.__init__(self)
        EventEmitter.__init__(self)
        self.client = client
        self.careful = client.careful
        self.method = None             # type: bytes
        self.uri = None                # type: bytes
        self.req_hdrs = None           # type: RawHeaderListType
        self.req_target = None         # type: bytes
        self.scheme = None             # type: bytes
        self.authority = None          # type: bytes
        self.res_version = None        # type: bytes
        self.tcp_conn = None           # type: TcpConnection
        self.origin = None             # type: OriginType
        self._conn_reusable = False
        self._req_body = False
        self._req_started = False
        self._retries = 0
        self._read_timeout_ev = None   # type: ScheduledEvent
        self._output_buffer = []       # type: list[bytes]

    def __repr__(self) -> str:
        status = [self.__class__.__module__ + "." + self.__class__.__name__]
        status.append('%s {%s}' % (self.method.decode('utf-8', 'replace') or "-",
                                   self.uri.decode('utf-8', 'replace') or "-"))
        if self.tcp_conn:
            status.append(self.tcp_conn.tcp_connected and 'connected' or 'disconnected')
        status.append(HttpMessageHandler.__repr__(self))
        return "<%s at %#x>" % (", ".join(status), id(self))

    def request_start(self, method: bytes, uri: bytes, req_hdrs: RawHeaderListType) -> None:
        """
        Start a request to uri using method, where req_hdrs is a list of (field_name, field_value)
        for the request headers.

        All values are bytes.
        """
        self.method = method
        self.uri = uri
        self.req_hdrs = req_hdrs
        try:
            self.origin = self._parse_uri(self.uri)
        except (TypeError, ValueError):
            return
        self.client.attach_conn(self.origin, self._handle_connect,
                                self._handle_connect_error, self.client.connect_timeout)
    # TODO: if we sent Expect: 100-continue, don't wait forever
    # (i.e., schedule something)

    def _parse_uri(self, uri: bytes) -> OriginType:
        """
        Given a bytes, parse out the host, port, authority and request target.
        Returns None if there is an error, otherwise the origin.
        """
        (scheme, authority, path, query, fragment) = urlsplit(uri)
        scheme = scheme.lower()
        if scheme == b'http':
            default_port = 80
        elif scheme == b'https':
            default_port = 443
        else:
            self.input_error(UrlError("Unsupported URL scheme '%s'" % \
                                      scheme.decode('utf-8', 'replace')))
            raise ValueError
        if b"@" in authority:
            userinfo, authority = authority.split(b"@", 1)
        if b":" in authority:
            host, port = authority.rsplit(b":", 1)
            try:
                port = int(port)
            except ValueError:
                self.input_error(UrlError("Non-integer port '%s' in URL" % \
                                          port.decode('utf-8', 'replace')))
                raise
            if not 1 <= port <= 65535:
                self.input_error(UrlError("URL port %i out of range" % port))
                raise ValueError
        else:
            host, port = authority, default_port
        if path == b"":
            path = b"/"
        self.scheme = scheme
        self.authority = authority
        self.req_target = urlunsplit((b'', b'', path, query, b''))
        return scheme, host, port

    def _req_start(self) -> None:
        """
        Queue the request headers for sending.
        """
        self._req_started = True
        req_hdrs = [i for i in self.req_hdrs if not i[0].lower() in req_rm_hdrs]
        req_hdrs.append((b"Host", self.authority))
        if self.client.idle_timeout > 0:
            req_hdrs.append((b"Connection", b"keep-alive"))
        else:
            req_hdrs.append((b"Connection", b"close"))
        if b"content-length" in header_names(req_hdrs):
            delimit = Delimiters.COUNTED
        elif self._req_body:
            req_hdrs.append((b"Transfer-Encoding", b"chunked"))
            delimit = Delimiters.CHUNKED
        else:
            delimit = Delimiters.NOBODY
        self._input_state = States.WAITING
        self.output_start(b"%s %s HTTP/1.1" % (self.method, self.req_target), req_hdrs, delimit)

    def request_body(self, chunk) -> None:
        "Send part of the request body. May be called zero to many times."
        if self._input_state == States.ERROR:
            return
        if not self._req_started:
            self._req_body = True
            self._req_start()
        self.output_body(chunk)

    def request_done(self, trailers: RawHeaderListType) -> None:
        """
        Signal the end of the request, whether or not there was a body. MUST
        be called exactly once for each request.
        """
        if self._input_state == States.ERROR:
            return
        if not self._req_started:
            self._req_start()
        self.output_end(trailers)

    def res_body_pause(self, paused: bool) -> None:
        "Temporarily stop / restart sending the response body."
        if self.tcp_conn and self.tcp_conn.tcp_connected:
            self.tcp_conn.pause(paused)

    # Methods called by tcp

    def _handle_connect(self, tcp_conn: TcpConnection) -> None:
        "The connection has succeeded."
        self.tcp_conn = tcp_conn
        self._set_read_timeout('connect')
        tcp_conn.on('data', self.handle_input)
        tcp_conn.on('close', self._conn_closed)
        tcp_conn.on('pause', self._req_body_pause)
        # FIXME: should this be done AFTER _req_start?
        self.output(b"") # kick the output buffer
        self.tcp_conn.pause(False)

    def _handle_connect_error(self, err_type, err_id, err_str) -> None:
        "The connection has failed."
        self.input_error(ConnectError(err_str))

    def _conn_closed(self) -> None:
        "The server closed the connection."
        self._clear_read_timeout()
        if self._input_buffer:
            self.handle_input(b"")
        if self._input_state == States.QUIET:
            pass # nothing to see here
        elif self._input_delimit == Delimiters.CLOSE:
            self.input_end([])
        elif self._input_state == States.WAITING:
            if self.method in idempotent_methods:
                if self._retries < self.client.retry_limit:
                    self.client.loop.schedule(self.client.retry_delay, self._retry)
                else:
                    self.input_error(
                        ConnectError("Tried to connect %s times." % (self._retries + 1)))
            else:
                self.input_error(ConnectError("Can't retry %s method" % \
                                              self.method.decode('utf-8', 'replace')))
        else:
            self.input_error(ConnectError(
                "Server dropped connection before the response was complete."))

    def _retry(self) -> None:
        "Retry the request."
        self._clear_read_timeout()
        self._retries += 1
        try:
            origin = self._parse_uri(self.uri)
        except (TypeError, ValueError):
            return
        self.client.attach_conn(origin, self._handle_connect,
                                self._handle_connect_error, self.client.connect_timeout)

    def _req_body_pause(self, paused: bool) -> None:
        "The client needs the application to pause/unpause the request body."
        self.emit('pause', paused)

    # Methods called by common.HttpMessageHandler

    def input_start(self, top_line: bytes, hdr_tuples: RawHeaderListType,
                    conn_tokens: List[bytes], transfer_codes: List[bytes],
                    content_length: int) -> Tuple[bool, bool]:
        """
        Take the top set of headers from the input stream, parse them
        and queue the request to be processed by the application.
        """
        self._clear_read_timeout()
        try:
            proto_version, status_txt = top_line.split(None, 1)
            proto, self.res_version = proto_version.rsplit(b'/', 1)
        except (ValueError, IndexError):
            self.input_error(StartLineError(top_line.decode('utf-8', 'replace')))
            raise ValueError
        if proto != b"HTTP" or self.res_version not in [b"1.0", b"1.1"]:
            self.input_error(HttpVersionError(proto_version.decode('utf-8', 'replace')))
            raise ValueError
        try:
            res_code, res_phrase = status_txt.split(None, 1)
        except ValueError:
            res_code = status_txt.rstrip()
            res_phrase = b""
        if b'close' not in conn_tokens:
            if (self.res_version == b"1.0" and b'keep-alive' in conn_tokens) \
              or self.res_version in [b"1.1"]:
                self._conn_reusable = True
        self._set_read_timeout('start')
        is_final = not res_code.startswith(b"1")
        allows_body = is_final and (res_code not in no_body_status) and (self.method != b"HEAD")
        if is_final:
            self.emit('response_start', res_code, res_phrase, hdr_tuples)
        else:
            self.emit('response_nonfinal', res_code, res_phrase, hdr_tuples)
        return allows_body, is_final

    def input_body(self, chunk: bytes) -> None:
        "Process a response body chunk from the wire."
        self._clear_read_timeout()
        self.emit('response_body', chunk)
        self._set_read_timeout('body')

    def input_end(self, trailers: RawHeaderListType) -> None:
        "Indicate that the response body is complete."
        self._clear_read_timeout()
        if self.tcp_conn:
            if self.tcp_conn.tcp_connected and self._conn_reusable:
                self.client.release_conn(self.tcp_conn, self.scheme)
            else:
                self.tcp_conn.close()
        self._dead_conn()
        self.tcp_conn = None
        self.emit('response_done', trailers)

    def input_error(self, err) -> None:
        "Indicate an error state."
        if err.client_recoverable and not self.careful:
            # This error isn't absolutely fatal, and we want to see the rest.
            # Still, we probably don't want to reuse this connection later.
            self._conn_reusable = False
        else:
            # It really is a fatal error.
            self._input_state = States.ERROR
            self._clear_read_timeout()
            self._dead_conn()
            if self.tcp_conn and self.tcp_conn.tcp_connected:
                self.tcp_conn.close()
            self.tcp_conn = None
        self.emit('error', err)

    def _dead_conn(self) -> None:
        "Inform the client that the connection is dead."
        self.client._dead_conn(self.origin)

    def output(self, chunk: bytes) -> None:
        self._output_buffer.append(chunk)
        if self.tcp_conn and self.tcp_conn.tcp_connected:
            self.tcp_conn.write(b"".join(self._output_buffer))
            self._output_buffer = []

    # misc

    def _set_read_timeout(self, kind) -> None:
        "Set the read timeout."
        if self.client.read_timeout:
            self._read_timeout_ev = self.client.loop.schedule(
                self.client.read_timeout, self.input_error, ReadTimeoutError(kind))

    def _clear_read_timeout(self) -> None:
        "Clear the read timeout."
        if self._read_timeout_ev:
            self._read_timeout_ev.delete()


def test_client(request_uri, out, err):  # pragma: no coverage
    "A simple demonstration of a client."
    from thor.loop import stop, run

    c = HttpClient()
    c.connect_timeout = 5
    c.careful = False
    x = c.exchange()

    @on(x)
    def response_start(status, phrase, headers):
        "Print the response headers."
        out(b"HTTP/%s %s %s\n" % (x.res_version, status, phrase))
        out(b"\n".join([b"%s:%s" % header for header in headers]))
        print()
        print()

    @on(x)
    def error(err_msg):
        if err_msg:
            err("\033[1;31m*** ERROR:\033[0;39m %s (%s)\n" % (err_msg.desc, err_msg.detail))
        if not err_msg.client_recoverable:
            stop()

    x.on('response_body', out)

    @on(x)
    def response_done(trailers):
        thor.schedule(1, stop)

    x.request_start(b"GET", request_uri, [])
    x.request_done([])
    run()


if __name__ == "__main__":
    import sys
    test_client(sys.argv[1].encode('ascii'), sys.stdout.buffer.write, sys.stderr.write)



#!/usr/bin/env python

import errno
import socket
import sys
import threading
import unittest


from thor import loop
from thor.udp import UdpEndpoint

test_host = "127.0.0.1"
test_port = 9002


class TestUdpEndpoint(unittest.TestCase):

    def setUp(self):
        self.loop = loop.make()
        self.ep1 = UdpEndpoint(self.loop)
        self.ep1.bind(test_host, test_port)
        self.ep1.on('datagram', self.input)
        self.ep1.pause(False)
        self.ep2 = UdpEndpoint()
        self.loop.schedule(5, self.timeout)
        self.timeout_hit = False
        self.datagrams = []

    def tearDown(self):
        self.ep1.shutdown()

    def timeout(self):
        self.timeout_hit = True
        self.loop.stop()

    def input(self, data, host, port):
        self.datagrams.append((data, host, port))

    def output(self, msg):
        self.ep2.send(msg, test_host, test_port)

    def test_basic(self):
        self.loop.schedule(1, self.output, b'foo!')
        self.loop.schedule(2, self.output, b'bar!')

        def check():
            self.assertEqual(self.datagrams[0][0], b'foo!')
            self.assertEqual(self.datagrams[1][0], b'bar!')
            self.loop.stop()
        self.loop.schedule(3, check)
        self.loop.run()

    def test_bigdata(self):
        self.loop.schedule(1, self.output, b'a' * 100)
        self.loop.schedule(2, self.output, b'b' * 1000)
        self.loop.schedule(3, self.output, b'c' * self.ep1.max_dgram)

        def check():
            self.assertEqual(self.datagrams[0][0], b'a' * 100)
            self.assertEqual(self.datagrams[1][0], b'b' * 1000)
            # we only check the first 1000 characters because, well,
            # it's lossy.
            self.assertEqual(self.datagrams[2][0][:1000], b'c' * 1000)
            self.loop.stop()
        self.loop.schedule(4, check)
        self.loop.run()

#   def test_pause(self):


if __name__ == '__main__':
    unittest.main()


#!/usr/bin/env python

import socket
import ssl
import sys
import threading
import unittest

from thor import loop
from thor.tls import TlsClient

# TODO: set up ssl servers for testing, don't use these.
test_host = b"www.google.com"
test_host2 = b"www.mnot.net"
test_port = 443

# TODO: update with framework
class TestTlsClientConnect(unittest.TestCase):

    def setUp(self):
        self.loop = loop.make()
        self.connect_count = 0
        self.error_count = 0
        self.last_error_type = None
        self.last_error = None
        self.timeout_hit = False
        self.conn = None
        def check_connect(conn):
            self.conn = conn
            self.assertTrue(conn.tcp_connected)
            self.connect_count += 1
            conn.write(b"GET / HTTP/1.0\r\n\r\n")
            conn.close()
            self.loop.schedule(1, self.loop.stop)
        def check_error(err_type, err_id, err_str):
            self.error_count += 1
            self.last_error_type = err_type
            self.last_error = err_id
            self.loop.schedule(1, self.loop.stop)
        def timeout():
            self.loop.stop()
            self.timeout_hit = True
        self.timeout = timeout
        self.client = TlsClient(self.loop)
        self.client.on('connect', check_connect)
        self.client.on('connect_error', check_error)

    def test_connect(self):
        self.client.connect(test_host, test_port)
        self.loop.schedule(2, self.timeout)
        self.loop.run()
        self.assertFalse(self.conn.tcp_connected)
        self.assertEqual(self.connect_count, 1)
        self.assertEqual(self.error_count, 0)
        self.assertEqual(self.timeout_hit, False)

# causing problems on CI infra
#    def test_connect_refused(self):
#        self.client.connect(test_host2, 25)
#        self.loop.schedule(3, self.timeout)
#        self.loop.run()
#        self.assertEqual(self.connect_count, 0)
#        self.assertEqual(self.error_count, 1)
#        self.assertEqual(self.last_error_type, socket.error)
#        self.assertEqual(self.last_error, ssl.errno.EINVAL)
#        self.assertEqual(self.timeout_hit, False)

    def test_connect_noname(self):
        self.client.connect('does.not.exist', test_port)
        self.loop.schedule(3, self.timeout)
        self.loop.run()
        self.assertEqual(self.connect_count, 0)
        self.assertEqual(self.error_count, 1)
        self.assertEqual(self.last_error_type, socket.gaierror)
        self.assertEqual(self.last_error, socket.EAI_NONAME)
        self.assertEqual(self.timeout_hit, False)

    def test_connect_timeout(self):
        self.client.connect('128.66.0.1', test_port, 1)
        self.loop.schedule(3, self.timeout)
        self.loop.run()
        self.assertEqual(self.connect_count, 0)
        self.assertEqual(self.error_count, 1)
        self.assertEqual(self.last_error_type, socket.error)
        self.assertEqual(self.last_error, ssl.errno.ETIMEDOUT)
        self.assertEqual(self.timeout_hit, False)

# TODO:
#   def test_pause(self):

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python

from __future__ import absolute_import
from thor.http.client import HttpClient
from thor.http.server import HttpServer
from thor.http.common import header_names, header_dict, get_header, \
  safe_methods, idempotent_methods, hop_by_hop_hdrs

debug = False



#!/usr/bin/env python

"""
Thor HTTP Server

This library allow implementation of an HTTP/1.1 server that is
"non-blocking," "asynchronous" and "event-driven" -- i.e., it achieves very
high performance and concurrency, so long as the application code does not
block (e.g., upon network, disk or database access). Blocking on one request
will block the entire server.

"""

import os
import sys
from typing import List, Tuple

from thor import schedule
from thor.events import EventEmitter, on
from thor.loop import LoopBase
from thor.tcp import TcpServer, TcpConnection

from thor.http.common import HttpMessageHandler, \
    States, Delimiters, \
    hop_by_hop_hdrs, \
    get_header, header_names, \
    RawHeaderListType
from thor.http.error import HttpVersionError, HostRequiredError, TransferCodeError


class HttpServer(EventEmitter):
    "An asynchronous HTTP server."

    tcp_server_class = TcpServer
    idle_timeout = 60 # in seconds

    def __init__(self, host: bytes, port: int, loop: LoopBase=None) -> None:
        EventEmitter.__init__(self)
        self.tcp_server = self.tcp_server_class(host, port, loop=loop)
        self.tcp_server.on('connect', self.handle_conn)
        schedule(0, self.emit, 'start')

    def handle_conn(self, tcp_conn: TcpConnection) -> None:
        http_conn = HttpServerConnection(tcp_conn, self)
        tcp_conn.on('data', http_conn.handle_input)
        tcp_conn.on('close', http_conn.conn_closed)
        tcp_conn.on('pause', http_conn.res_body_pause)
        tcp_conn.pause(False)

    def shutdown(self) -> None:
        "Stop the server"
        # TODO: Finish outstanding requests w/ timeout?
        self.tcp_server.shutdown()
        self.emit('stop')


class HttpServerConnection(HttpMessageHandler, EventEmitter):
    "A handler for an HTTP server connection."
    default_state = States.WAITING

    def __init__(self, tcp_conn: TcpConnection, server: HttpServer) -> None:
        HttpMessageHandler.__init__(self)
        EventEmitter.__init__(self)
        self.tcp_conn = tcp_conn
        self.server = server
        self.ex_queue = []  # type: list[HttpServerExchange] # queue of exchanges
        self.output_paused = False

    def req_body_pause(self, paused: bool) -> None:
        """
        Indicate that the server should pause (True) or unpause (False) the
        request.
        """
        self.tcp_conn.pause(paused)

    # Methods called by tcp

    def res_body_pause(self, paused: bool) -> None:
        "Pause/unpause sending the response body."
        self.output_paused = paused
        self.emit('pause', paused)
        if not paused:
            self.drain_exchange_queue()

    def conn_closed(self) -> None:
        "The server connection has closed."
#        for exchange in self.ex_queue:
#            exchange.pause() # FIXME - maybe a connclosed err?
        self.ex_queue = []
        self.tcp_conn = None

    # Methods called by common.HttpRequestHandler

    def output(self, data: bytes) -> None:
        self.tcp_conn.write(data)

    def input_start(self, top_line: bytes, hdr_tuples: RawHeaderListType,
                    conn_tokens: List[bytes], transfer_codes: List[bytes],
                    content_length: int) -> Tuple[bool, bool]:
        """
        Take the top set of headers from the input stream, parse them
        and queue the request to be processed by the application.
        """
        try:
            method, req_line = top_line.split(None, 1)
            uri, req_version = req_line.rsplit(None, 1)
            req_version = req_version.rsplit(b'/', 1)[1]
        except (ValueError, IndexError):
            self.input_error(HttpVersionError(top_line.decode('utf-8', 'replace')))
            # TODO: more fine-grained
            raise ValueError
        if b'host' not in header_names(hdr_tuples):
            self.input_error(HostRequiredError())
            raise ValueError
        for code in transfer_codes:
            # we only support 'identity' and chunked' codes in requests
            if code not in [b'identity', b'chunked']:
                self.input_error(TransferCodeError(code.decode('utf-8', 'replace')))
                raise ValueError
        exchange = HttpServerExchange(self, method, uri, hdr_tuples, req_version)
        self.ex_queue.append(exchange)
        self.server.emit('exchange', exchange)
        if not self.output_paused:
            # we only start new requests if we have some output buffer
            # available.
            exchange.request_start()
        allows_body = bool(content_length and content_length > 0) or (transfer_codes != [])
        return allows_body, True

    def input_body(self, chunk: bytes) -> None:
        "Process a request body chunk from the wire."
        self.ex_queue[-1].emit('request_body', chunk)

    def input_end(self, trailers: RawHeaderListType) -> None:
        "Indicate that the request body is complete."
        self.ex_queue[-1].emit('request_done', trailers)

    def input_error(self, err) -> None:
        """
        Indicate a parsing problem with the request body (which
        hasn't been queued as an exchange yet).
        """
        if err.server_recoverable:
            self.emit('error', err)
        else:
            self._input_state = States.ERROR
            status_code, status_phrase = err.server_status or (b"500", b'Internal Server Error')
            hdrs = [(b'Content-Type', b'text/plain'),]
            body = err.desc.encode("utf-8")
            if err.detail:
                body += b" (%s)" % err.detail.encode("utf-8")
            ex = HttpServerExchange(self, b'', b'', [], b"1.1")
            ex.response_start(status_code, status_phrase, hdrs)
            ex.response_body(body)
            ex.response_done([])
            self.ex_queue.append(ex)

# FIXME: connection?

#        if self.tcp_conn and not err.server_recoverable:
#            self.tcp_conn.close()
#            self.tcp_conn = None

# TODO: if in mid-request, we need to send an error event and clean up.
#        self.ex_queue[-1].emit('error', err)

    def drain_exchange_queue(self) -> None:
        """
        Walk through the exchange queue and kick off unstarted requests
        until we run out of output buffer.
        """
        # TODO: probably have a separate metric for outstanding requests,
        # rather than just the write queue size.
        for exchange in self.ex_queue:
            if not exchange.started:
                exchange.request_start()


class HttpServerExchange(EventEmitter):
    """
    A request/response interaction on an HTTP server.
    """

    def __init__(self, http_conn, method: bytes, uri: bytes, req_hdrs: RawHeaderListType,
                 req_version: bytes) -> None:
        EventEmitter.__init__(self)
        self.http_conn = http_conn
        self.method = method
        self.uri = uri
        self.req_hdrs = req_hdrs
        self.req_version = req_version
        self.started = False

    def __repr__(self) -> str:
        status = [self.__class__.__module__ + "." + self.__class__.__name__]
        status.append('%s {%s}' % (self.method or "-", self.uri or "-"))
        return "<%s at %#x>" % (", ".join(status), id(self))

    def request_start(self) -> None:
        self.started = True
        self.emit('request_start', self.method, self.uri, self.req_hdrs)

    def response_start(self, status_code: bytes, status_phrase: bytes,
                       res_hdrs: RawHeaderListType) -> None:
        "Start a response. Must only be called once per response."
        res_hdrs = [i for i in res_hdrs if not i[0].lower() in hop_by_hop_hdrs]
        try:
            body_len = int(get_header(res_hdrs, b"content-length").pop(0))
        except (IndexError, ValueError):
            body_len = None
        if body_len is not None:
            delimit = Delimiters.COUNTED
            res_hdrs.append((b"Connection", b"keep-alive"))
        elif self.req_version == b"1.1":
            delimit = Delimiters.CHUNKED
            res_hdrs.append((b"Transfer-Encoding", b"chunked"))
        else:
            delimit = Delimiters.CLOSE
            res_hdrs.append((b"Connection", b"close"))

        self.http_conn.output_start(
            b"HTTP/1.1 %s %s" % (status_code, status_phrase), res_hdrs, delimit)

    def response_body(self, chunk: bytes) -> None:
        "Send part of the response body. May be called zero to many times."
        self.http_conn.output_body(chunk)

    def response_done(self, trailers: RawHeaderListType) -> None:
        """
        Signal the end of the response, whether or not there was a body. MUST
        be called exactly once for each response.
        """
        self.http_conn.output_end(trailers)


def test_handler(x): # pragma: no cover
    @on(x, 'request_start')
    def go(*args):
        print("start: %s on %s" % (str(args[1]), id(x.http_conn)))
        x.response_start(b'200', b"OK", [])
        x.response_body(b'foo!')
        x.response_done([])

    @on(x, 'request_body')
    def body(chunk):
        print("body: %s" % chunk)

    @on(x, 'request_done')
    def done(trailers):
        print("done: %s" % str(trailers))


if __name__ == "__main__":
    from thor.loop import run
    sys.stderr.write("PID: %s\n" % os.getpid())
    h, p = b'127.0.0.1', int(sys.argv[1])
    demo_server = HttpServer(h, p)
    demo_server.on('exchange', test_handler)
    run()



#!/usr/bin/env python

"""
Event utilities, including:

* EventEmitter - in the style of Node.JS.
* on - a decorator for making functions and methods listen to events.
"""

from collections import defaultdict
from typing import Any, Callable, Dict, List


class EventEmitter(object):
    """
    An event emitter, in the style of Node.JS.
    """

    def __init__(self) -> None:
        self.__events = defaultdict(list)  # type: Dict[str, List[Callable]]
        self.__sink = None                 # type: object

    def __getstate__(self) -> Dict[str, Any]:
        state = self.__dict__.copy()
        del state["_EventEmitter__events"]
        return state

    def on(self, event: str, listener: Callable) -> None:
        """
        Call listener when event is emitted.
        """
        self.__events[event].append(listener)
        self.emit('newListener', event, listener)

    def once(self, event: str, listener: Callable) -> None:
        """
        Call listener the first time event is emitted.
        """
        def mycall(*args):
            listener(*args)
            self.removeListener(event, mycall)
        self.on(event, mycall)

    def removeListener(self, event: str, listener: Callable) -> None:
        """
        Remove a specific listener from an event.

        If called for a specific listener by a previous listener
        for the same event, that listener will not be fired.
        """
        self.__events.get(event, [listener]).remove(listener)

    def removeListeners(self, *events: str) -> None:
        """
        Remove all listeners from an event; if no event
        is specified, remove all listeners for all events.

        If called from an event listener, other listeners
        for that event will still be fired.
        """
        if events:
            for event in events:
                self.__events[event] = []
        else:
            self.__events = defaultdict(list)

    def listeners(self, event: str) -> List[Callable]:
        """
        Return a list of listeners for an event.
        """
        return self.__events.get(event, [])

    def events(self) -> List[str]:
        """
        Return a list of events being listened for.
        """
        return list(self.__events)

    def emit(self, event: str, *args) -> None:
        """
        Emit the event (with any given args) to
        its listeners.
        """
        events = self.__events.get(event, [])
        if len(events):
            for e in events:
                e(*args)
        else:
            sink_event = getattr(self.__sink, event, None)
            if sink_event:
                sink_event(*args)

    def sink(self, sink: object) -> None:
        """
        If no listeners are found for an event, call
        the method that shares the event's name (if present)
        on the event sink.
        """
        self.__sink = sink

    # TODO: event bubbling


def on(obj: object, event: str=None) -> Callable:
    """
    Decorator to call a function when an object emits
    the specified event.
    """
    def wrap(funk):
        obj.on(event or funk.__name__, funk)
        return funk
    return wrap



#!/usr/bin/env python

try:
    import SocketServer
except ImportError:
    import socketserver as SocketServer

import errno
import socket
import sys
import threading
import unittest

from thor import loop
from thor.tcp import TcpClient

test_host = "127.0.0.1"
test_port = 9002

class LittleRequestHandler(SocketServer.BaseRequestHandler):
    def handle(self):
        # Echo the back to the client
        data = self.request.recv(1024)
        self.request.send(data)

class LittleServer(SocketServer.ThreadingMixIn, SocketServer.TCPServer):
    allow_reuse_address = True

# TODO: update with framework
class TestTcpClientConnect(unittest.TestCase):

    def setUp(self):
        self.loop = loop.make()
        self.connect_count = 0
        self.error_count = 0
        self.last_error_type = None
        self.last_error = None
        self.timeout_hit = False
        self.conn = None
        def check_connect(conn):
            self.conn = conn
            self.assertTrue(conn.tcp_connected)
            self.connect_count += 1
            conn.write(b"test")
            conn.close()
            self.loop.schedule(1, self.loop.stop)
        def check_error(err_type, err_id, err_str):
            self.error_count += 1
            self.last_error_type = err_type
            self.last_error = err_id
            self.loop.schedule(1, self.loop.stop)
        def timeout():
            self.loop.stop()
            self.timeout_hit = True
        self.timeout = timeout
        self.client = TcpClient(self.loop)
        self.client.on('connect', check_connect)
        self.client.on('connect_error', check_error)

    def test_connect(self):
        self.server = LittleServer(
            (test_host, test_port),
            LittleRequestHandler
        )
        t = threading.Thread(target=self.server.serve_forever)
        t.setDaemon(True)
        t.start()
        self.client.connect(test_host, test_port)
        self.loop.schedule(2, self.timeout)
        self.loop.run()
        self.assertFalse(self.conn.tcp_connected)
        self.assertEqual(self.connect_count, 1)
        self.assertEqual(self.error_count, 0)
        self.assertEqual(self.timeout_hit, False)
        self.server.shutdown()
        self.server.socket.close()

    def test_connect_refused(self):
        self.client.connect(test_host, test_port + 1)
        self.loop.schedule(3, self.timeout)
        self.loop.run()
        self.assertEqual(self.connect_count, 0)
        self.assertEqual(self.error_count, 1)
        self.assertEqual(self.last_error_type, socket.error)
        self.assertEqual(self.last_error, errno.ECONNREFUSED)
        self.assertEqual(self.timeout_hit, False)

    def test_connect_noname(self):
        self.client.connect('does.not.exist', test_port)
        self.loop.schedule(3, self.timeout)
        self.loop.run()
        self.assertEqual(self.connect_count, 0)
        self.assertEqual(self.error_count, 1)
        self.assertEqual(self.last_error_type, socket.gaierror)
        self.assertEqual(self.last_error, socket.EAI_NONAME)
        self.assertEqual(self.timeout_hit, False)

    def test_connect_timeout(self):
        self.client.connect('128.66.0.1', test_port, 1)
        self.loop.schedule(3, self.timeout)
        self.loop.run()
        self.assertEqual(self.connect_count, 0)
        self.assertEqual(self.error_count, 1)
        self.assertEqual(self.last_error_type, socket.error)
        self.assertEqual(self.last_error, errno.ETIMEDOUT, 
                         errno.errorcode.get(self.last_error, self.last_error))
        self.assertEqual(self.timeout_hit, False)

# TODO:
#   def test_pause(self):

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python

"""
push-based asynchronous TCP

This is a generic library for building event-based / asynchronous
TCP servers and clients.

It uses a push model; i.e., the network connection pushes data to
you (using a 'data' event), and you push data to the network connection
(using the write method).
"""

import errno
import os
import sys
import socket
from typing import Tuple # pylint: disable=unused-import

from thor.loop import EventSource, LoopBase, schedule, ScheduledEvent


class TcpConnection(EventSource):
    """
    An asynchronous TCP connection.

    Emits:
     - data (chunk): incoming data
     - close (): the other party has closed the connection
     - pause (bool): whether the connection has been paused

    It will emit the 'data' even every time incoming data is
    available;

    > def process(data):
    >   print "got some data:", data
    > tcp_conn.on('data', process)

    When you want to write to the connection, just write to it:

    > tcp_conn.write(data)

    If you want to close the connection from your side, just call close:

    > tcp_conn.close()

    Note that this will flush any data already written.

    If the other side closes the connection, The 'close' event will be
    emitted;

    > def handle_close():
    >   print "oops, they don't like us any more..."
    > tcp_conn.on('close', handle_close)

    If you write too much data to the connection and the buffers fill up,
    pause_cb will be emitted with True to tell you to stop sending data
    temporarily;

    > def handle_pause(paused):
    >   if paused:
    >       # stop sending data
    >   else:
    >       # it's OK to start again
    > tcp_conn.on('pause', handle_pause)

    Note that this is advisory; if you ignore it, the data will still be
    buffered, but the buffer will grow.

    Likewise, if you want to pause the connection because your buffers
    are full, call pause;

    > tcp_conn.pause(True)

    but don't forget to tell it when it's OK to send data again;

    > tcp_conn.pause(False)

    NOTE that connections are paused to start with; if you want to start
    getting data from them, you'll need to pause(False).
    """

    # TODO: play with various buffer sizes
    write_bufsize = 16
    read_bufsize = 1024 * 16

    block_errs = set([errno.EAGAIN, errno.EWOULDBLOCK, errno.ETIMEDOUT])
    close_errs = set([
        errno.EBADF, errno.ECONNRESET, errno.ESHUTDOWN,
        errno.ECONNABORTED, errno.ECONNREFUSED,
        errno.ENOTCONN, errno.EPIPE])

    def __init__(self, sock: socket.socket, host: bytes, port: int, loop: LoopBase=None) -> None:
        EventSource.__init__(self, loop)
        self.socket = sock
        self.host = host
        self.port = port
        self.tcp_connected = True # we assume a connected socket
        self._input_paused = True # we start with input paused
        self._output_paused = False
        self._closing = False
        self._write_buffer = []   # type: list[bytes]

        self.register_fd(sock.fileno())
        self.on('fd_readable', self.handle_readable)
        self.on('fd_writable', self.handle_writable)
        self.on('fd_close', self._handle_close)

    def __repr__(self) -> str:
        status = [self.__class__.__module__ + "." + self.__class__.__name__]
        status.append(self.tcp_connected and 'connected' or 'disconnected')
        status.append('%s:%s' % (self.host, self.port))
        if self._input_paused:
            status.append('input paused')
        if self._output_paused:
            status.append('output paused')
        if self._closing:
            status.append('closing')
        if self._write_buffer:
            status.append('%s write buffered' % len(self._write_buffer))
        return "<%s at %#x>" % (", ".join(status), id(self))

    def handle_readable(self) -> None:
        "The connection has data read for reading"
        try:
            data = self.socket.recv(self.read_bufsize)
        except (socket.error, OSError) as why:
            if why.args[0] in self.block_errs:
                return
            elif why.args[0] in self.close_errs:
                self._handle_close()
                return
            else:
                raise
        if data == b"":
            self._handle_close()
        else:
            self.emit('data', data)

    def handle_writable(self) -> None:
        "The connection is ready for writing; write any buffered data."
        if len(self._write_buffer) > 0:
            data = b"".join(self._write_buffer)
            try:
                sent = self.socket.send(data)
            except (socket.error, OSError) as why:
                if why.args[0] in self.block_errs:
                    return
                elif why.args[0] in self.close_errs:
                    self._handle_close()
                    return
                else:
                    raise
            if sent < len(data):
                self._write_buffer = [data[sent:]]
            else:
                self._write_buffer = []
        if self._output_paused and len(self._write_buffer) < self.write_bufsize:
            self._output_paused = False
            self.emit('pause', False)
        if self._closing:
            self._close()
        if len(self._write_buffer) == 0:
            self.event_del('fd_writable')

    def write(self, data: bytes) -> None:
        "Write data to the connection."
        self._write_buffer.append(data)
        if len(self._write_buffer) > self.write_bufsize:
            self._output_paused = True
            self.emit('pause', True)
        self.event_add('fd_writable')

    def pause(self, paused: bool) -> None:
        """
        Temporarily stop/start reading from the connection and pushing
        it to the app.
        """
        if paused:
            self.event_del('fd_readable')
        else:
            self.event_add('fd_readable')
        self._input_paused = paused

    def close(self) -> None:
        "Flush buffered data (if any) and close the connection."
        self.pause(True)
        if len(self._write_buffer) > 0:
            self._closing = True
        else:
            self._close()
        # TODO: should loop stop automatically close all conns?

    def _handle_close(self) -> None:
        "The connection has been closed by the other side."
        self._close()
        self.emit('close')

    def _close(self) -> None:
        self.tcp_connected = False
        self.removeListeners('fd_readable', 'fd_writable', 'fd_close')
        self.unregister_fd()
        if self.socket:
            self.socket.close()


class TcpServer(EventSource):
    """
    An asynchronous TCP server.

    Emits:
      - connect (tcp_conn): upon connection

    To start listening:

    > s = TcpServer(host, port)
    > s.on('connect', conn_handler)

    conn_handler is called every time a new client connects.
    """
    def __init__(self, host: bytes, port: int, sock: socket.socket=None,
                 loop: LoopBase=None) -> None:
        EventSource.__init__(self, loop)
        self.host = host
        self.port = port
        self.sock = sock or server_listen(host, port)
        self.on('fd_readable', self.handle_accept)
        self.register_fd(self.sock.fileno(), 'fd_readable')
        schedule(0, self.emit, 'start')

    def handle_accept(self) -> None:
        try:
            conn, addr = self.sock.accept()
        except (TypeError, IndexError):
            # sometimes accept() returns None if we have
            # multiple processes listening
            return
        conn.setblocking(False)
        tcp_conn = TcpConnection(conn, self.host, self.port, self._loop)
        self.emit('connect', tcp_conn)

    # TODO: should loop stop close listening sockets?

    def shutdown(self) -> None:
        "Stop accepting requests and close the listening socket."
        self.removeListeners('fd_readable')
        self.sock.close()
        self.emit('stop')
        # TODO: emit close?


def server_listen(host: bytes, port: int, backlog: int=None) -> socket.socket:
    "Return a socket listening to host:port."
    # TODO: IPV6
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.setblocking(False)
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    sock.bind((host, port))
    sock.listen(backlog or socket.SOMAXCONN)
    return sock


class TcpClient(EventSource):
    """
    An asynchronous TCP client.

    Emits:
      - connect (tcp_conn): upon connection
      - connect_error (err_type, err_id, err_str): if there's a problem
        before getting a connection. err_type is socket.error or
        socket.gaierror; err_id is the specific error encountered, and
        err_str is its textual description.

    To connect to a server:

    > c = TcpClient()
    > c.on('connect', conn_handler)
    > c.on('connect_error', error_handler)
    > c.connect(host, port)

    conn_handler will be called with the tcp_conn as the argument
    when the connection is made.
    """
    def __init__(self, loop: LoopBase=None) -> None:
        EventSource.__init__(self, loop)
        self.host = None  # type: bytes
        self.port = None  # type: int
        self._timeout_ev = None  # type: ScheduledEvent
        self._error_sent = False
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.sock.setblocking(False)
        self.on('fd_error', self.handle_conn_error)
        self.register_fd(self.sock.fileno(), 'fd_writable')
        self.event_add('fd_error')

    def connect(self, host: bytes, port: int, connect_timeout: float=None) -> None:
        """
        Connect to host:port (with an optional connect timeout)
        and emit 'connect' when connected, or 'connect_error' in
        the case of an error.
        """
        self.host = host
        self.port = port
        self.on('fd_writable', self.handle_connect)
        # TODO: use socket.getaddrinfo(); needs to be non-blocking.
        try:
            err = self.sock.connect_ex((host, port))
        except socket.gaierror as why:
            self.handle_conn_error(socket.gaierror, why)
            return
        except socket.error as why:
            self.handle_conn_error(socket.error, why)
            return
        if err != errno.EINPROGRESS:
            self.handle_conn_error(socket.error, socket.error(err, os.strerror(err)))
            return
        if connect_timeout:
            self._timeout_ev = self._loop.schedule(
                connect_timeout,
                self.handle_conn_error,
                socket.error,
                socket.error(errno.ETIMEDOUT, os.strerror(errno.ETIMEDOUT)),
                True)

    def handle_connect(self) -> None:
        self.unregister_fd()
        if self._timeout_ev:
            self._timeout_ev.delete()
        if self._error_sent:
            return
        err = self.sock.getsockopt(socket.SOL_SOCKET, socket.SO_ERROR)
        if err:
            self.handle_conn_error(socket.error, (err, os.strerror(err)))
        else:
            tcp_conn = TcpConnection(self.sock, self.host, self.port, self._loop)
            self.emit('connect', tcp_conn)

    def handle_conn_error(self, err_type=None, why=None, close: bool=True) -> None:
        """
        Handle a connect error.

        @err_type - e.g., socket.error; defaults to socket.error
        @why - tuple of [err_id, err_str]
        @close - whether the error means the socket should be closed
        """
        if self._timeout_ev:
            self._timeout_ev.delete()
        if self._error_sent:
            return
        if err_type is None:
            err_type = socket.error
            err_id = self.sock.getsockopt(socket.SOL_SOCKET, socket.SO_ERROR)
            err_str = os.strerror(err_id)
        else:
            if hasattr(why, 'args'):
                args = why.args
            else:
                args = why
            err_id = args[0]
            err_str = args[1]
        self._error_sent = True
        self.unregister_fd()
        self.emit('connect_error', err_type, err_id, err_str)
        if close:
            self.sock.close()


if __name__ == "__main__":
    # quick demo server
    from thor.loop import run, stop
    server = TcpServer(b'localhost', int(sys.argv[-1]))
    def handle_conn(conn):
        conn.pause(False)
        def echo(chunk):
            if chunk.strip().lower() in ['quit', 'stop']:
                stop()
            else:
                conn.write("-> %s" % chunk)
        conn.on('data', echo)
    server.on('connect', handle_conn)
    run()



#!/usr/bin/env python

"""
push-based asynchronous UDP

This is a generic library for building event-based / asynchronous
UDP servers and clients.
"""

import errno
import socket

from thor.loop import EventSource, LoopBase


class UdpEndpoint(EventSource):
    """
    An asynchronous UDP endpoint.

    Emits:
      - datagram (data, address): upon recieving a datagram.

    To start:

    > s = UdpEndpoint(host, port)
    > s.on('datagram', datagram_handler)
    """
    recv_buffer = 8192
    _block_errs = set([errno.EAGAIN, errno.EWOULDBLOCK])

    def __init__(self, loop: LoopBase=None) -> None:
        EventSource.__init__(self, loop)
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.sock.setblocking(False)
        self.max_dgram = min((2**16 - 40), self.sock.getsockopt(
            socket.SOL_SOCKET, socket.SO_SNDBUF
        ))
        self.on('fd_readable', self.handle_datagram)
        self.register_fd(self.sock.fileno())

    def __repr__(self) -> str:
        status = [self.__class__.__module__ + "." + self.__class__.__name__]
        return "<%s at %#x>" % (", ".join(status), id(self))

    def bind(self, host: str, port: int) -> None:
        """
        Bind the socket bound to host:port. If called, must be before
        sending or receiving.

        Can raise socket.error if binding fails.
        """
        # TODO: IPV6
        self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.sock.bind((host, port))

    def shutdown(self) -> None:
        "Close the listening socket."
        self.removeListeners('fd_readable')
        self.sock.close()
        # TODO: emit close?

    def pause(self, paused: bool) -> None:
        "Control incoming datagram events."
        if paused:
            self.event_del('fd_readable')
        else:
            self.event_add('fd_readable')

    def send(self, datagram: bytes, host: str, port: int) -> None:
        "send datagram to host:port."
        try:
            self.sock.sendto(datagram, (host, port))
        except socket.error as why:
            if why in self._block_errs:
                pass # we drop these on the floor. It's UDP, after all.
            else:
                raise

    def handle_datagram(self) -> None:
        "Handle an incoming datagram, emitting the 'datagram' event."
        # TODO: consider pre-allocating buffers.
        # TODO: is it best to loop here?
        while True:
            try:
                data, addr = self.sock.recvfrom(self.recv_buffer)
            except socket.error as why:
                if why.args[0] in self._block_errs:
                    break
                else:
                    raise
            self.emit('datagram', data, addr[0], addr[1])



##!/usr/bin/env python

import socket
import sys
import time
import unittest

import framework

import thor
from thor.events import on
from thor.http import HttpServer

class TestHttpServer(framework.ClientServerTestCase):

    def create_server(self, test_host, test_port, server_side):
        server = HttpServer(test_host, test_port, loop=self.loop)
        server_side(server)
        @on(self.loop)
        def stop():
            server.shutdown()

    def create_client(self, test_host, test_port, client_side):
        def run_client(client_side1):
            client = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            client.connect((test_host, test_port))
            client_side1(client)
            client.close()
        self.move_to_thread(target=run_client, args=(client_side,))

    def check_exchange(self, exchange, expected):
        """
        Given an exchange, check that the status, phrase and body are as
        expected, and verify that it actually happened.
        """
        exchange.test_happened = False

        @on(exchange)
        def error(err_msg):
            exchange.test_happened = True
            self.assertEqual(err_msg, expected.get('error', err_msg))
            self.loop.stop()

        @on(exchange)
        def request_start(method, uri, headers):
            self.assertEqual(method, expected.get('method', method))
            self.assertEqual(uri, expected.get('phrase', uri))

        exchange.tmp_req_body = b""
        @on(exchange)
        def request_body(chunk):
            exchange.tmp_req_body += chunk

        @on(exchange)
        def request_done(trailers):
            exchange.test_happened = True
            self.assertEqual(
                trailers,
                expected.get('req_trailers', trailers)
            )
            self.assertEqual(
                exchange.tmp_req_body,
                expected.get('body', exchange.tmp_req_body)
            )
            self.loop.stop()

        @on(self.loop)
        def stop():
            self.assertTrue(exchange.test_happened)


    def test_basic(self):
        def server_side(server):
            def check(exchange):
                self.check_exchange(exchange, {
                    'method': b'GET',
                    'uri': b'/'
                })
            server.on('exchange', check)

        def client_side(client_conn):
            client_conn.sendall(b"""\
GET / HTTP/1.1
Host: %s:%i

""" % (framework.test_host, framework.test_port))
            time.sleep(1)
            client_conn.close()
        self.go([server_side], [client_side])


    def test_extraline(self):
        def server_side(server):
            def check(exchange):
                self.check_exchange(exchange, {
                    'method': b'GET',
                    'uri': b'/'
                })
            server.on('exchange', check)

        def client_side(client_conn):
            client_conn.sendall(b"""\

GET / HTTP/1.1
Host: %s:%i\r
\r
""" % (framework.test_host, framework.test_port))
            time.sleep(1)
            client_conn.close()
        self.go([server_side], [client_side])


    def test_post(self):
        def server_side(server):
            def check(exchange):
                self.check_exchange(exchange, {
                    'method': b'POST',
                    'uri': b'/foo'
                })
            server.on('exchange', check)

        def client_side(client_conn):
            client_conn.sendall(b"""\
POST / HTTP/1.1
Host: %s:%i
Content-Type: text/plain
Content-Length: 5

12345""" % (framework.test_host, framework.test_port))
            time.sleep(1)
            client_conn.close()
        self.go([server_side], [client_side])


    def test_post_extra_crlf(self):
        def server_side(server):
            def check(exchange):
                self.check_exchange(exchange, {
                    'method': b'POST',
                    'uri': b'/foo'
                })
            server.on('exchange', check)

        def client_side(client_conn):
            client_conn.sendall(b"""\
POST / HTTP/1.1
Host: %s:%i
Content-Type: text/plain
Content-Length: 5

12345
""" % (framework.test_host, framework.test_port))
            time.sleep(1)
            client_conn.close()
        self.go([server_side], [client_side])


#    def test_pipeline(self):
#        def server_side(server):
#            server.ex_count = 0
#            def check(exchange):
#                self.check_exchange(exchange, {
#                    'method': 'GET',
#                    'uri': '/'
#                })
#                server.ex_count += 1
#            server.on('exchange', check)
#            @on(self.loop)
#            def stop():
#                self.assertEqual(server.ex_count, 2)
#
#        def client_side(client_conn):
#            client_conn.sendall("""\
#GET / HTTP/1.1
#Host: %s:%i
#
#GET / HTTP/1.1
#Host: %s:%i
#
#""" % (
#    framework.test_host, framework.test_port,
#    framework.test_host, framework.test_port
#))
#            time.sleep(1)
#            client_conn.close()
#        self.go([server_side], [client_side])



#    def test_conn_close(self):
#    def test_req_nobody(self):
#    def test_res_nobody(self):
#    def test_bad_http_version(self):
#    def test_pause(self):
#    def test_extra_crlf_after_post(self):
#    def test_absolute_uri(self): # ignore host header
#    def test_host_header(self):#
#    def test_unknown_transfercode(self): # should be 501
#    def test_shutdown(self):
#    def test_alternate_tcp_server(self):
#    def test_startline_encoding(self):


if __name__ == '__main__':
    unittest.main()




#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
import unittest

from framework import DummyHttpParser
from thor.http.common import Delimiters

import thor.http.error as error


class TestHttpParser(unittest.TestCase):

    def setUp(self):
        self.parser = DummyHttpParser()

    def checkSingleMsg(self, inputs, body, expected_err=None, close=False):
        """
        Check a single HTTP message.
        """
        assert type(inputs) == type([])
        for chunk in inputs:
            self.parser.handle_input(chunk % {
                b'body': body,
                b'body_len': len(body)
            })
        states = self.parser.test_states

        if not expected_err:
            self.assertFalse("ERROR" in states, self.parser.test_err)
            self.assertTrue(states.count('START') == 1, states)
            self.assertTrue(states.index('START') < states.index('BODY'))
            if close:
                self.assertEqual(self.parser._input_delimit, Delimiters.CLOSE)
            else:
                self.assertTrue(states.index('END') + 1 == len(states))
            self.assertEqual(body, self.parser.test_body, "%s not equal to %s" % (body[:20], self.parser.test_body[:20]))
        else:
            self.assertTrue("ERROR" in states, states)
            self.assertEqual(self.parser.test_err.__class__, expected_err)

    def checkMultiMsg(self, inputs, body, count):
        """
        Check pipelined messages. Assumes the same body for each (for now).
        """
        for chunk in inputs:
            self.parser.handle_input(chunk % {
                b'body': body,
                b'body_len': len(body)
            })
        states = self.parser.test_states
        self.assertFalse("ERROR" in self.parser.test_states, self.parser.test_err)
        self.parser.check(self, {'states': ['START', 'BODY', 'END'] * count})

    def test_hdrs(self):
        body = b"12345678901234567890"
        self.checkSingleMsg([b"""\
http/1.1 200 OK
Content-Type: text/plain
Foo: bar
Content-Length: %(body_len)i
Foo: baz, bam

%(body)s"""], body)
        self.parser.check(self, {
            'hdrs': [
                (b'Content-Type', b" text/plain"),
                (b'Foo', b" bar"),
                (b'Content-Length', b" %i" % len(body)),
                (b'Foo', b" baz, bam"),
            ]
        })

    def test_hdrs_nocolon(self):
        body = b"12345678901234567890"
        self.checkSingleMsg([b"""\
http/1.1 200 OK
Content-Type: text/plain
Foo bar
Content-Length: %(body_len)i

%(body)s"""], body)
        # FIXME: error?

    def test_hdr_case(self):
        body = b"12345678901234567890"
        self.checkSingleMsg([b"""\
http/1.1 200 OK
Content-Type: text/plain
content-LENGTH: %(body_len)i

%(body)s"""], body)

    def test_hdrs_whitespace_before_colon(self):
        body = b"lorum ipsum whatever goes after that."
        self.checkSingleMsg([b"""\
HTTP/1.1 200 OK
Content-Type: text/plain
Content-Length : %(body_len)i

%(body)s"""], body, error.HeaderSpaceError)

    def test_hdrs_fold(self):
        body = b"lorum ipsum whatever goes after that."
        self.checkSingleMsg([b"""\
HTTP/1.1 200 OK
Content-Type: text/plain
Foo: bar
     baz
Content-Length: %(body_len)i

%(body)s"""], body)
        foo_val = [v for k,v in self.parser.test_hdrs if k == b'Foo'][-1]
        self.assertEqual(foo_val, b" bar baz")
        headers = [k for k,v in self.parser.test_hdrs]
        self.assertEqual(headers, [b'Content-Type', b'Foo', b'Content-Length'])

    def test_hdrs_noname(self):
        body = b"lorum ipsum whatever goes after that."
        self.checkSingleMsg([b"""\
HTTP/1.1 200 OK
Content-Type: text/plain
: bar
Content-Length: %(body_len)i

%(body)s"""], body)
        headers = [k for k,v in self.parser.test_hdrs]
        self.assertEqual(headers, [b'Content-Type', b'', b'Content-Length'])


    def test_hdrs_utf8(self):
        body = b"lorum ipsum whatever goes after that."
        self.checkSingleMsg([u"""\
HTTP/1.1 200 OK
Content-Type: text/plain
Foo: ედუარდ შევარდნაძე
Content-Length: %(body_len)i

%(body)s""".encode('utf-8')], body)
        foo_val = [v for k,v in self.parser.test_hdrs if k == b'Foo'][-1]
        self.assertEqual(foo_val.decode('utf-8'), u" ედუარდ შევარდნაძე")

    def test_hdrs_null(self):
        body = b"lorum ipsum whatever goes after that."
        self.checkSingleMsg([u"""\
HTTP/1.1 200 OK
Content-Type: text/plain
Foo: \0
Content-Length: %(body_len)i

%(body)s""".encode('utf-8')], body)
        foo_val = [v for k,v in self.parser.test_hdrs if k == b'Foo'][-1]
        self.assertEqual(foo_val, b" \0")


    def test_cl_delimit_11(self):
        body = b"lorum ipsum whatever goes after that."
        self.checkSingleMsg([b"""\
HTTP/1.1 200 OK
Content-Type: text/plain
Content-Length: %(body_len)i

%(body)s"""], body)

    def test_cl_delimit_10(self):
        body = b"abcdefghijklmnopqrstuvwxyz"
        self.checkSingleMsg([b"""\
HTTP/1.0 200 OK
Content-Type: text/plain
Content-Length: %(body_len)i

%(body)s"""], body)

    def test_close_delimit(self):
        body = b"abcdefghijklmnopqrstuvwxyz"
        self.checkSingleMsg([b"""\
HTTP/1.0 200 OK
Content-Type: text/plain

%(body)s"""], body, close=True)

    def test_extra_line(self):
        body = b"lorum ipsum whatever goes after that."
        self.checkSingleMsg([b"""\

HTTP/1.1 200 OK
Content-Type: text/plain
Content-Length: %(body_len)i

%(body)s"""], body)

    def test_extra_lines(self):
        body = b"lorum ipsum whatever goes after that."
        self.checkSingleMsg([b"""\



HTTP/1.1 200 OK
Content-Type: text/plain
Content-Length: %(body_len)i

%(body)s"""], body)

    def test_telnet_client(self):
        body = "lorum ipsum whatever goes after that."
        self.checkSingleMsg([a.encode('ascii') for a in """\



HTTP/1.1 200 OK
Content-Type: text/plain
Content-Length: %(body_len)i

%(body)s""" % {'body': body, 'body_len': len(body)}], body.encode('ascii'))


    def test_naughty_first_header(self):
        body = b"lorum ipsum whatever goes after that."
        self.checkSingleMsg([b"""\
HTTP/1.1 200 OK
    Content-Type: text/plain
Content-Length: %(body_len)i

%(body)s"""], body, error.TopLineSpaceError)

    def test_cl_header_case(self):
        body = b"12345678901234567890"
        self.checkSingleMsg([b"""\
HTTP/1.1 200 OK
Content-Type: text/plain
content-LENGTH: %(body_len)i

%(body)s"""], body)

    def test_chunk_delimit(self):
        body = b"aaabbbcccdddeeefffggghhhiii"
        self.checkSingleMsg([b"""\
HTTP/1.1 200 OK
Content-Type: text/plain
Transfer-Encoding: chunked

%(body_len)x\r
%(body)s\r
0\r
\r
"""], body)

    def test_chunk_exact(self):
        body = b"aaabbbcccdddeeefffggghhhiii"
        self.checkSingleMsg([b"""\
HTTP/1.1 200 OK
Content-Type: text/plain
Transfer-Encoding: chunked

""", b"""\
%(body_len)x\r
%(body)s\r
""", b"""\
0\r
\r
"""], body)

    def test_chunk_split(self):
        body = b"aaabbbcccdddeeefffggghhhiii"
        self.checkSingleMsg([b"""\
HTTP/1.1 200 OK
Content-Type: text/plain
Transfer-Encoding: chunked

""", b"""\
%(body_len)x\r
%(body)s\r
0""", b"""\
\r
Foo: bar\r
\r
"""], body)

    def test_chunk_exact_offset(self):
        body = b"aaabbbcccdddeeefffggghhhiii"
        self.checkSingleMsg([b"""\
HTTP/1.1 200 OK
Content-Type: text/plain
Transfer-Encoding: chunked

""", b"""\
%(body_len)x\r
%(body)s""", b"""\r
0\r
\r
"""], body)
    def test_chunk_more(self):
        body = b"1234567890"
        self.checkSingleMsg([b"""\
HTTP/1.1 200 OK
Content-Type: text/plain
Transfer-Encoding: chunked

""", b"""\
%(body_len)x\r
%(body)s\r
%(body_len)x\r
%(body)s\r
0\r
\r
""" % {b'body': body, b'body_len': len(body)}], body * 2)


    def test_transfer_case(self):
        body = b"aaabbbcccdddeeefffggghhhiii"
        self.checkSingleMsg([b"""\
HTTP/1.1 200 OK
Content-Type: text/plain
Transfer-Encoding: cHuNkEd

%(body_len)x\r
%(body)s\r
0\r
\r
"""], body)

    def test_big_chunk(self):
        body = b"aaabbbcccdddeeefffggghhhiii" * 1000000
        self.checkSingleMsg([b"""\
HTTP/1.1 200 OK
Content-Type: text/plain
Transfer-Encoding: chunked

%(body_len)x\r
%(body)s\r
0\r
\r
"""], body)

    def xxxtest_small_chunks(self):
        num_chunks = 10000
        body = b"a" * num_chunks
        inputs = [b"""\
HTTP/1.1 200 OK
Content-Type: text/plain
Transfer-Encoding: chunked

"""]
        for i in range(num_chunks):
            inputs.append(b"""\
1\r
a\r
""")
        inputs.append(b"""\
0\r
\r
""")
        self.checkSingleMsg(inputs, body)

    def test_split_chunk(self):
        body = b"abcdefg123456"
        self.checkSingleMsg([b"""\
HTTP/1.1 200 OK
Content-Type: text/plain
Transfer-Encoding: chunked

%(body_len)x\r
abcdefg""",
b"""\
123456\r
0\r
\r
"""], body)

    def test_split_chunk_length(self):
        body = b"do re mi so fa la ti do"
        self.checkSingleMsg([b"""\
HTTP/1.1 200 OK
Content-Type: text/plain
Transfer-Encoding: chunked

%(body_len)x""",
b"""\
\r
%(body)s\r
0\r
\r
"""], body)

    def test_chunk_bad_syntax(self):
        body = b"abc123def456ghi789"
        self.checkSingleMsg([b"""\
HTTP/1.1 200 OK
Content-Type: text/plain
Transfer-Encoding: chunked

ZZZZ\r
%(body)s\r
0\r
\r
"""], body, error.ChunkError)

    def test_chunk_nonfinal(self):
        body = b"abc123def456ghi789"
        self.checkSingleMsg([b"""\
HTTP/1.1 200 OK
Content-Type: text/plain
Transfer-Encoding: chunked, foo

%(body)s"""], body, close=True)

    def test_cl_dup(self):
        body = b"abc123def456ghi789"
        self.checkSingleMsg([b"""\
HTTP/1.1 200 OK
Content-Type: text/plain
Content-Length: %(body_len)i
Content-Length: %(body_len)i

%(body)s"""], body)

    def test_cl_conflict(self):
        body = b"abc123def456ghi789"
        self.checkSingleMsg([b"""\
HTTP/1.1 200 OK
Content-Type: text/plain
Content-Length: 2
Content-Length: %(body_len)i

%(body)s"""], body, error.DuplicateCLError)

    def test_cl_bad_syntax(self):
        body = b"abc123def456ghi789"
        self.checkSingleMsg([b"""\
HTTP/1.1 200 OK
Content-Type: text/plain
Content-Length: 2abc

%(body)s"""], body, error.MalformedCLError)

    def test_chunk_ext(self):
        body = b"abc123def456ghi789"
        self.checkSingleMsg([b"""\
HTTP/1.1 200 OK
Content-Type: text/plain
Transfer-Encoding: chunked

%(body_len)x; myext=foobarbaz\r
%(body)s\r
0\r
\r
"""], body)

    def test_trailers(self):
        body = b"abc123def456ghi789"
        self.checkSingleMsg([b"""\
HTTP/1.1 200 OK
Content-Type: text/plain
Transfer-Encoding: chunked

%(body_len)x\r
%(body)s\r
0\r
Foo: bar
Baz: 1
\r
"""], body)
        self.assertEqual(self.parser.test_trailers,
            [(b'Foo', b' bar'), (b'Baz', b' 1')]
        )

    def test_pipeline_chunked(self):
        body = b"abc123def456ghi789"
        self.checkMultiMsg([b"""\
HTTP/1.1 200 OK
Content-Type: text/plain
Transfer-Encoding: chunked

%(body_len)x\r
%(body)s\r
0\r
\r
HTTP/1.1 404 Not Found
Content-Type: text/plain
Transfer-Encoding: chunked

%(body_len)x\r
%(body)s\r
0\r
\r
"""], body, 2)

    def test_pipeline_cl(self):
        body = b"abc123def456ghi789"
        self.checkMultiMsg([b"""\
HTTP/1.1 200 OK
Content-Type: text/plain
Content-Length: %(body_len)i

%(body)sHTTP/1.1 404 Not Found
Content-Type: text/plain
Content-Length: %(body_len)i

%(body)s"""], body, 2)

# TODO:
#    def test_nobody_delimit(self):
#    def test_pipeline_nobody(self):
#    def test_chunked_then_length(self):
#    def test_length_then_chunked(self):


if __name__ == '__main__':
    unittest.main()


#!/usr/bin/env python

import socket
import sys
import time
import unittest

import framework

import thor
from thor.events import on

class TestTcpServer(framework.ClientServerTestCase):

    def create_server(self, host, port, server_side):
        server = thor.TcpServer(host, port, loop=self.loop)
        server.conn_count = 0
        def run_server(conn):
            server.conn_count += 1
            server_side(conn)
        server.on('connect', run_server)
        @on(self.loop)
        def stop():
            self.assertTrue(server.conn_count > 0)
            server.shutdown()

    def create_client(self, host, port, client_side):
        def run_client(client_side1):
            client = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            client.connect((host, port))
            client_side1(client)
            client.close()
        self.move_to_thread(target=run_client, args=(client_side,))
        self.loop.schedule(1, self.loop.stop)

    def test_basic(self):
        def server_side(server_conn):
            self.server_recv = 0
            def check_data(chunk):
                self.assertEqual(chunk, b"foo!")
                self.server_recv += 1
            server_conn.on('data', check_data)
            server_conn.pause(False)
            server_conn.write(b"bar!")

        def client_side(client_conn):
            sent = client_conn.send(b'foo!')

        self.go([server_side], [client_side])
        self.assertTrue(self.server_recv > 0, self.server_recv)

# TODO:
#   def test_pause(self):
#   def test_shutdown(self):

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python

"""
Asynchronous event loops

This is a generic library for building asynchronous event loops, using
Python's built-in poll / epoll / kqueue support.
"""

import select
import sys
import time as systime
from typing import Callable, List, Set, Tuple

from thor.events import EventEmitter


__all__ = ['run', 'stop', 'schedule', 'time', 'running', 'debug']


class EventSource(EventEmitter):
    """
    Base class for objects that the loop will direct interesting
    events to.

    An instance should map to one thing with an interesting file
    descriptor, registered with register_fd.
    """
    def __init__(self, loop: 'LoopBase'=None) -> None:
        EventEmitter.__init__(self)
        self._loop = loop or _loop
        self._interesting_events = set()  # type: set[str]
        self._fd = None  # type: int

    def register_fd(self, fd: int, event: str=None) -> None:
        """
        Register myself with the loop using file descriptor fd.
        If event is specified, start emitting it.
        """
        self._fd = fd
        self._loop.register_fd(self._fd, [], self)
        self.event_add(event)

    def unregister_fd(self) -> None:
        "Unregister myself from the loop."
        if self._fd:
            self._loop.unregister_fd(self._fd)
            self._fd = None

    def event_add(self, event: str) -> None:
        "Start emitting the given event."
        if event and event not in self._interesting_events:
            self._interesting_events.add(event)
            self._loop.event_add(self._fd, event)

    def event_del(self, event: str) -> None:
        "Stop emitting the given event."
        if event in self._interesting_events:
            self._interesting_events.remove(event)
            self._loop.event_del(self._fd, event)

    def interesting_events(self) -> Set[str]:
        return self._interesting_events


class LoopBase(EventEmitter):
    """
    Base class for async loops.
    """
    _event_types = {}   # type: Dict[int, str] # map of event types to names; override.

    def __init__(self, precision: float=None) -> None:
        EventEmitter.__init__(self)
        self.precision = precision or .5 # of running scheduled queue (secs)
        self.running = False      # whether or not the loop is running (read-only)
        self.__sched_events = []  # type: list[Tuple[float, Callable]]
        self._fd_targets = {}     # type: dict[int, EventEmitter]
        self.__now = None         # type: float
        self._eventlookup = dict([(v, k) for (k, v) in self._event_types.items()])
        self.__event_cache = {}   # type: dict[int, set[str]]

    def __repr__(self) -> str:
        name = self.__class__.__name__
        running = self.running and "running" or "not-running"
        events = len(self.__sched_events)
        targets = len(self._fd_targets)
        return "<%s - %s, %i events, %i fd_targets>" % (name, running, events, targets)

    def run(self) -> None:
        "Start the loop."
        self.running = True
        last_event_check = 0  # type: float
        self.__now = systime.time()
        self.emit('start')
        while self.running:
            if debug:
                fd_start = systime.time()
            self._run_fd_events()
            self.__now = systime.time()
            if debug:
                delay = self.__now - fd_start
                if delay >= self.precision * 1.5:
                    sys.stderr.write("WARNING: long fd delay (%.2f)\n" % delay)
            # find scheduled events
            delay = self.__now - last_event_check
            if delay >= self.precision * 0.90:
                if debug:
                    if last_event_check and (delay >= self.precision * 4):
                        sys.stderr.write("WARNING: long loop delay (%.2f)\n" % delay)
                    if len(self.__sched_events) > 5000:
                        sys.stderr.write("WARNING: %i events scheduled\n" % \
                                         len(self.__sched_events))
                last_event_check = self.__now
                for event in self.__sched_events:
                    when, what = event
                    if self.running and self.__now >= when:
                        try:
                            self.__sched_events.remove(event)
                        except ValueError:
                            # a previous event may have removed this one.
                            continue
                        if debug:
                            ev_start = systime.time()
                        what()
                        if debug:
                            delay = systime.time() - ev_start
                            if delay > self.precision * 2:
                                sys.stderr.write("WARNING: long event delay (%.2f): %s\n" % \
                                                 (delay, repr(what)))
                    else:
                        break

    def _run_fd_events(self) -> None:
        "Run loop-specific FD events."
        raise NotImplementedError

    def stop(self) -> None:
        "Stop the loop and unregister all fds."
        self.__sched_events = []
        self.__now = None
        self.running = False
        for fd in list(self._fd_targets):
            self.unregister_fd(fd)
        self.emit('stop')

    def register_fd(self, fd: int, events: List[str], target) -> None:
        "emit events on target when they occur on fd."
        raise NotImplementedError

    def unregister_fd(self, fd: int) -> None:
        "Stop emitting events from fd."
        raise NotImplementedError

    def fd_count(self) -> int:
        "Return how many FDs are currently monitored by the loop."
        return len(self._fd_targets)

    def event_add(self, fd: int, event: str) -> None:
        "Start emitting event for fd."
        raise NotImplementedError

    def event_del(self, fd: int, event: str) -> None:
        "Stop emitting event for fd"
        raise NotImplementedError

    def _fd_event(self, event: str, fd: int) -> None:
        "An event has occured on an fd."
        if fd in self._fd_targets:
            self._fd_targets[fd].emit(event)
        # TODO: automatic unregister on 'close'?

    def time(self) -> float:
        "Return the current time (to avoid a system call)."
        return self.__now or systime.time()

    def schedule(self, delta: float, callback: Callable, *args) -> 'ScheduledEvent':
        """
        Schedule callable callback to be run in delta seconds with *args.

        Returns an object which can be used to later remove the event, by
        calling its delete() method.
        """
        def cb(): # FIXME: can't compare functions in py3. Suck.
            if callback:
                callback(*args)
        cb.__name__ = callback.__name__
        new_event = (self.time() + delta, cb)
        events = self.__sched_events
        self._insort(events, new_event)
        return ScheduledEvent(self, new_event)

    def schedule_del(self, event):
        try:
            self.__sched_events.remove(event)
        except ValueError: # already gone
            pass

    @staticmethod
    def _insort(a, x, lo=0, hi=None):
        if lo < 0:
            raise ValueError('lo must be non-negative')
        if hi is None:
            hi = len(a)
        while lo < hi:
            mid = (lo+hi)//2
            if x[0] < a[mid][0]:
                hi = mid
            else: lo = mid+1
        a.insert(lo, x)

    def _eventmask(self, events: List[str]) -> int:
        "Calculate the mask for a list of events."
        eventmask = 0
        for event in events:
            eventmask |= self._eventlookup.get(event, 0)
        return eventmask

    def _filter2events(self, evfilter: int) -> Set[str]:
        "Calculate the events implied by a given filter."
        if evfilter not in self.__event_cache:
            events = set()
            for et in self._event_types:
                if et & evfilter:
                    events.add(self._event_types[et])
            self.__event_cache[evfilter] = events
        return self.__event_cache[evfilter]


class ScheduledEvent(object):
    """
    Holds a scheduled event.
    """
    def __init__(self, loop: LoopBase, event: Tuple[float, Callable]) -> None:
        self._loop = loop
        self._event = event
        self._deleted = False

    def delete(self) -> None:
        if not self._deleted:
            self._loop.schedule_del(self._event)
            self._deleted = True


class PollLoop(LoopBase):
    """
    A poll()-based async loop.
    """

    def __init__(self, *args) -> None:
        # pylint: disable=E1101
        self._event_types = {
            select.POLLIN: 'fd_readable',
            select.POLLOUT: 'fd_writable',
            select.POLLERR: 'fd_error',
            select.POLLHUP: 'fd_close'}
    #        select.POLLNVAL - TODO

        LoopBase.__init__(self, *args)
        self._poll = select.poll()
        # pylint: enable=E1101

    def register_fd(self, fd, events, target):
        self._fd_targets[fd] = target
        self._poll.register(fd, self._eventmask(events))

    def unregister_fd(self, fd):
        self._poll.unregister(fd)
        del self._fd_targets[fd]

    def event_add(self, fd, event):
        eventmask = self._eventmask(self._fd_targets[fd].interesting_events())
        self._poll.register(fd, eventmask)

    def event_del(self, fd, event):
        eventmask = self._eventmask(self._fd_targets[fd].interesting_events())
        self._poll.register(fd, eventmask)

    def _run_fd_events(self):
        event_list = self._poll.poll(self.precision)
        for fileno, eventmask in event_list:
            for event in self._filter2events(eventmask):
                self._fd_event(event, fileno)


class EpollLoop(LoopBase):
    """
    An epoll()-based async loop.
    """

    def __init__(self, *args):
        # pylint: disable=E1101
        self._event_types = {
            select.EPOLLIN: 'fd_readable',
            select.EPOLLOUT: 'fd_writable',
            select.EPOLLHUP: 'fd_close',
            select.EPOLLERR: 'fd_error'
        }
        LoopBase.__init__(self, *args)
        self._epoll = select.epoll()
        # pylint: enable=E1101

    def register_fd(self, fd, events, target):
        eventmask = self._eventmask(events)
        if fd in self._fd_targets:
            self._epoll.modify(fd, eventmask)
        else:
            self._fd_targets[fd] = target
            self._epoll.register(fd, eventmask)

    def unregister_fd(self, fd):
        self._epoll.unregister(fd)
        del self._fd_targets[fd]

    def event_add(self, fd, event):
        eventmask = self._eventmask(self._fd_targets[fd].interesting_events())
        self._epoll.modify(fd, eventmask)

    def event_del(self, fd, event):
        try:
            eventmask = self._eventmask(self._fd_targets[fd].interesting_events())
        except KeyError:
            return # no longer interested
        self._epoll.modify(fd, eventmask)

    def _run_fd_events(self):
        event_list = self._epoll.poll(self.precision)
        for fileno, eventmask in event_list:
            for event in self._filter2events(eventmask):
                self._fd_event(event, fileno)


class KqueueLoop(LoopBase):
    """
    A kqueue()-based async loop.
    """
    def __init__(self, *args):
        self._event_types = {
            select.KQ_FILTER_READ: 'fd_readable',
            select.KQ_FILTER_WRITE: 'fd_writable'}
        LoopBase.__init__(self, *args)
        self.max_ev = 50 # maximum number of events to pull from the queue
        self._kq = select.kqueue()

    # TODO: override schedule() to use kqueue event scheduling.

    def register_fd(self, fd, events, target):
        self._fd_targets[fd] = target
        for event in events:
            self.event_add(fd, event)

    def unregister_fd(self, fd):
        try:
            obj = self._fd_targets[fd]
        except KeyError:
            return
        for event in list(obj.interesting_events()):
            obj.event_del(event)
        del self._fd_targets[fd]

    def event_add(self, fd, event):
        eventmask = self._eventmask([event])
        if eventmask:
            ev = select.kevent(fd, eventmask, select.KQ_EV_ADD | select.KQ_EV_ENABLE)
            self._kq.control([ev], 0, 0)

    def event_del(self, fd, event):
        eventmask = self._eventmask([event])
        if eventmask:
            ev = select.kevent(fd, eventmask, select.KQ_EV_DELETE)
            self._kq.control([ev], 0, 0)

    def _run_fd_events(self):
        events = self._kq.control([], self.max_ev, self.precision)
        for e in events:
            event_types = self._filter2events(e.filter)
            for event_type in event_types:
                self._fd_event(event_type, int(e.ident))
            if e.flags & select.KQ_EV_EOF:
                self._fd_event('fd_close', int(e.ident))
            if e.flags & select.KQ_EV_ERROR:
                pass
            # TODO: pull errors, etc. out of flags and fflags
            #   If the read direction of the socket has shutdown, then
    		#	the filter also sets EV_EOF in flags, and returns the
    		#	socket error (if any) in fflags.  It is possible for
    		#	EOF to be returned (indicating the connection is gone)
    		#	while there is still data pending in the socket
    		#	buffer.


def make(precision=None) -> LoopBase:
    """
    Create and return a named loop that is suitable for the current system. If
    _precision_ is given, it indicates how often scheduled events will be run.

    Returned loop instances have all of the methods and instance variables
    that *thor.loop* has.
    """
    if hasattr(select, 'epoll'):
        loop = EpollLoop(precision)
    elif hasattr(select, 'kqueue'):
        loop = KqueueLoop(precision) # type: ignore
    elif hasattr(select, 'poll'):
        loop = PollLoop(precision)   # type: ignore
    else:
        raise ImportError("What is this thing, a Windows box?")
    return loop

_loop = make() # by default, just one big loop.
run = _loop.run
stop = _loop.stop
schedule = _loop.schedule
time = _loop.time
running = _loop.running
debug = False



#!/usr/bin/env python

"""
Thor HTTP Errors
"""

from typing import Tuple # pylint: disable=unused-import


class HttpError(Exception):
    desc = u"Unknown Error"
    server_status = None # type: Tuple[bytes, bytes]  # status this produces in a server
    server_recoverable = False # whether a server can recover the connection
    client_recoverable = False # whether a client can recover the connection

    def __init__(self, detail: str=None) -> None:
        Exception.__init__(self)
        self.detail = detail

    def __repr__(self) -> str:
        status = [self.__class__.__module__ + "." + self.__class__.__name__]
        return "<%s at %#x>" % (", ".join(status), id(self))

# General parsing errors

class ChunkError(HttpError):
    desc = u"Chunked encoding error"

class DuplicateCLError(HttpError):
    desc = u"Duplicate Content-Length header"
    server_status = (b"400", b"Bad Request")
    client_recoverable = True

class MalformedCLError(HttpError):
    desc = u"Malformed Content-Length header"
    server_status = (b"400", b"Bad Request")

class ExtraDataError(HttpError):
    desc = u"Extra data was sent after this message was supposed to end"

class StartLineError(HttpError):
    desc = u"The start line of the message couldn't be parsed"

class HttpVersionError(HttpError):
    desc = u"Unrecognised HTTP version"
    server_status = (b"505", b"HTTP Version Not Supported")

class ReadTimeoutError(HttpError):
    desc = u"Read Timeout"

class TransferCodeError(HttpError):
    desc = u"Unknown request transfer coding"
    server_status = (b"501", b"Not Implemented")

class HeaderSpaceError(HttpError):
    desc = u"Whitespace at the end of a header field-name"
    server_status = (b"400", b"Bad Request")
    client_recoverable = True

class TopLineSpaceError(HttpError):
    desc = u"Whitespace after top line, before first header"
    server_status = (b"400", b"Bad Request")
    client_recoverable = True

class TooManyMsgsError(HttpError):
    desc = u"Too many messages to parse"
    server_status = (b"400", b"Bad Request")

# client-specific errors

class UrlError(HttpError):
    desc = u"Unsupported or invalid URI"
    server_status = (b"400", b"Bad Request")

class LengthRequiredError(HttpError):
    desc = u"Content-Length required"
    server_status = (b"411", b"Length Required")
    client_recoverable = True

class ConnectError(HttpError):
    desc = u"Connection error"
    server_status = (b"504", b"Gateway Timeout")

# server-specific errors

class HostRequiredError(HttpError):
    desc = u"Host header required"
    server_recoverable = True



#!/usr/bin/env python
# coding: utf-8

try:
    import SocketServer
except ImportError:
    import socketserver as SocketServer

import sys
import time
import unittest

import framework
test_host = framework.test_host
test_port = framework.test_port


import thor
from thor.events import on
from thor.http import HttpClient

thor.loop.debug = True

class LittleServer(SocketServer.ThreadingMixIn, SocketServer.TCPServer):
    allow_reuse_address = True


class TestHttpClient(framework.ClientServerTestCase):

    def create_server(self, test_host, test_port, server_side):
        class LittleRequestHandler(SocketServer.BaseRequestHandler):
            handle = server_side
        server = LittleServer((framework.test_host, framework.test_port), LittleRequestHandler)
        self.move_to_thread(target=server.serve_forever)

        @on(self.loop)
        def stop():
            server.shutdown()
            server.server_close()

    def create_client(self, host, port, client_side):
        client = HttpClient(loop=self.loop)
        client.connect_timeout = 1
        client_side(client)

    def check_exchange(self, exchange, expected):
        """
        Given an exchange, check that the status, phrase and body are as
        expected, and verify that it actually happened.
        """
        exchange.test_happened = False

        @on(exchange)
        def error(err_msg):
            exchange.test_happened = True
            self.assertEqual(err_msg, expected.get('error', err_msg))

        @on(exchange)
        def response_start(status, phrase, headers):
            self.assertEqual(exchange.res_version, expected.get('version', exchange.res_version))
            self.assertEqual(status, expected.get('status', status))
            self.assertEqual(phrase, expected.get('phrase', phrase))

        exchange.tmp_res_body = b""
        @on(exchange)
        def response_body(chunk):
            exchange.tmp_res_body += chunk

        @on(exchange)
        def response_done(trailers):
            exchange.test_happened = True
            self.assertEqual(exchange.tmp_res_body, expected.get('body', exchange.tmp_res_body))

        @on(self.loop)
        def stop():
            self.assertTrue(exchange.test_happened, expected)



    def test_basic(self):
        def client_side(client):
            exchange = client.exchange()
            self.check_exchange(exchange, {
                'version': b"1.1",
                'status': b"200",
                'phrase': b'OK',
                'body': b"12345"
            })

            @on(exchange)
            def response_done(trailers):
                self.loop.stop()

            req_uri = b"http://%s:%i/basic" % (test_host, test_port)
            exchange.request_start(b"GET", req_uri, [])
            exchange.request_done([])

        def server_side(conn):
            conn.request.send(b"""\
HTTP/1.1 200 OK
Content-Type: text/plain
Content-Length: 5
Connection: close

12345""")
            conn.request.close()
        self.go([server_side], [client_side])


    def test_chunked_response(self):
        def client_side(client):
            exchange = client.exchange()
            self.check_exchange(exchange, {
                'version': b"1.1",
                'status': b"200",
                'phrase': b'OK',
                'body': b"12345"
            })
            @on(exchange)
            def response_done(trailers):
                self.loop.stop()

            req_uri = b"http://%s:%i/chunked_response" % (test_host, test_port)
            exchange.request_start(b"GET", req_uri, [])
            exchange.request_done([])

        def server_side(conn):
            conn.request.send(b"""\
HTTP/1.1 200 OK
Content-Type: text/plain
Transfer-Encoding: chunked

5\r
12345\r
0\r
\r
""")
            conn.request.close()
        self.go([server_side], [client_side])


    def test_chunked_request(self):
        req_body = b"54321"
        def client_side(client):
            exchange = client.exchange()
            self.check_exchange(exchange, {
                'version': b"1.1",
                'status': b"200",
                'phrase': b'OK',
                'body': b"12345"
            })
            @on(exchange)
            def response_done(trailers):
                self.loop.stop()

            req_uri = b"http://%s:%i/chunked_request" % (test_host, test_port)
            exchange.request_start(b"POST", req_uri, [])
            exchange.request_body(req_body)
            exchange.request_body(req_body)
            exchange.request_done([])

        def server_side(conn):
            conn.request.send(b"""\
HTTP/1.1 200 OK
Content-Type: text/plain
Transfer-Encoding: chunked

5\r
12345\r
0\r
\r
""")
            # TODO: check server-side recv
            conn.request.close()
        self.go([server_side], [client_side])


    def test_1xx(self):
        req_body = b"54321"
        def client_side(client):
            self.nonfinal_seen = False
            exchange = client.exchange()
            self.check_exchange(exchange, {
                'version': b"1.1",
                'status': b"200",
                'phrase': b'OK',
                'body': b"12345"
            })
            @on(exchange)
            def response_nonfinal(status_code, status_phrase, headers):
                self.nonfinal_seen = True
            @on(exchange)
            def response_done(trailers):
                self.assertTrue(self.nonfinal_seen)
                self.loop.stop()

            req_uri = b"http://%s:%i/chunked_request" % (test_host, test_port)
            exchange.request_start(b"POST", req_uri, [])
            exchange.request_body(req_body)
            exchange.request_body(req_body)
            exchange.request_done([])

        def server_side(conn):
            conn.request.send(b"""\
HTTP/1.1 110 Whatever
This: that

HTTP/1.1 200 OK
Content-Type: text/plain
Transfer-Encoding: chunked

5\r
12345\r
0\r
\r
""")
            # TODO: check server-side recv
            conn.request.close()
        self.go([server_side], [client_side])


    def test_multiconn(self):
        self.test_req_count = 0
        def check_done(trailers):
            self.test_req_count += 1
            if self.test_req_count == 2:
                self.loop.stop()

        def client_side(client):
            exchange1 = client.exchange()
            self.check_exchange(exchange1, {
                'version': b"1.1",
                'status': b"200",
                'phrase': b'OK',
                'body': b"12345"
            })

            exchange1.on('response_done', check_done)
            exchange2 = client.exchange()
            self.check_exchange(exchange2, {
                'version': b"1.1",
                'status': b"200",
                'phrase': b'OK',
                'body': b"12345"
            })
            exchange2.on('response_done', check_done)

            req_uri = b"http://%s:%i/multiconn" % (test_host, test_port)
            exchange1.request_start(b"GET", req_uri, [])
            exchange2.request_start(b"GET", req_uri, [])
            exchange1.request_done([])
            exchange2.request_done([])

        def server_side(conn):
            conn.request.send(b"""\
HTTP/1.1 200 OK
Content-Type: text/plain
Content-Length: 5
Connection: close

12345""")
            time.sleep(1)
            conn.request.close()
        self.go([server_side], [client_side])


    def test_conn_refuse_err(self):
        def client_side(client):
            exchange = client.exchange()
            @on(exchange)
            def error(err_msg):
                self.assertEqual(err_msg.__class__, thor.http.error.ConnectError)
                self.loop.stop()

            req_uri = b"http://%s:%i/conn_refuse_err" % (test_host, test_port)
            exchange.request_start(b"GET", req_uri, [])
            exchange.request_done([])
        self.go([], [client_side])


    # FIXME: works because dns is currently blocking
    def test_conn_noname_err(self):
        client = HttpClient(loop=self.loop)
        exchange = client.exchange()
        @on(exchange)
        def error(err_msg):
            self.assertEqual(err_msg.__class__, thor.http.error.ConnectError)
            self.loop.stop()

        req_uri = b"http://foo.bar/conn_noname_err"
        exchange.request_start(b"GET", req_uri, [])
        exchange.request_done([])


    def test_url_err(self):
        client = HttpClient(loop=self.loop)
        exchange = client.exchange()
        @on(exchange)
        def error(err_msg):
            self.assertEqual(err_msg.__class__, thor.http.error.UrlError)
            self.loop.stop()

        req_uri = b"foo://%s:%i/url_err" % (test_host, test_port)
        exchange.request_start(b"GET", req_uri, [])
        exchange.request_done([])


    def test_url_port_err(self):
        client = HttpClient(loop=self.loop)
        exchange = client.exchange()
        @on(exchange)
        def error(err_msg):
            self.assertEqual(err_msg.__class__, thor.http.error.UrlError)
            self.loop.stop()

        req_uri = b"http://%s:ABC123/url_port_err" % (test_host)
        exchange.request_start(b"GET", req_uri, [])
        exchange.request_done([])


    def test_url_port_range(self):
        client = HttpClient(loop=self.loop)
        exchange = client.exchange()
        @on(exchange)
        def error(err_msg):
            self.assertEqual(err_msg.__class__, thor.http.error.UrlError)
            self.loop.stop()

        req_uri = b"http://%s:80000/" % (test_host)
        exchange.request_start(b"GET", req_uri, [])
        exchange.request_done([])


    def test_http_version_err(self):
        def client_side(client):
            exchange = client.exchange()
            @on(exchange)
            def error(err_msg):
                self.assertEqual(err_msg.__class__, thor.http.error.HttpVersionError)
                self.loop.stop()

            req_uri = b"http://%s:%i/http_version_err" % (test_host, test_port)
            exchange.request_start(b"GET", req_uri, [])
            exchange.request_done([])

        def server_side(conn):
            conn.request.send(b"""\
HTTP/2.5 200 OK
Content-Type: text/plain
Content-Length: 5
Connection: close

12345""")
            conn.request.close()
        self.go([server_side], [client_side])


    def test_http_start_encoding(self):
        def client_side(client):
            exchange = client.exchange()
            self.check_exchange(exchange, {
                'version': b"1.1",
                'status': b"200",
                'phrase': u'ÖK'.encode('utf-8'),
                'body': b"12345"
            })
            @on(exchange)
            def response_done(trailers):
                self.loop.stop()

            req_uri = b"http://%s:%i/http_start_encoding" % (test_host, test_port)
            exchange.request_start(b"GET", req_uri, [])
            exchange.request_done([])

        def server_side(conn):
            conn.request.send(b"""\
HTTP/1.1 200 \xc3\x96K
Content-Type: text/plain
Content-Length: 5
Connection: close

12345""")
            conn.request.close()
        self.go([server_side], [client_side])


    def test_http_protoname_err(self):
        def client_side(client):
            exchange = client.exchange()
            @on(exchange)
            def error(err_msg):
                self.assertEqual(err_msg.__class__, thor.http.error.HttpVersionError)
                self.loop.stop()

            req_uri = b"http://%s:%i/protoname_err" % (test_host, test_port)
            exchange.request_start(b"GET", req_uri, [])
            exchange.request_done([])

        def server_side(conn):
            conn.request.send(b"""\
ICY/1.1 200 OK
Content-Type: text/plain
Content-Length: 5
Connection: close

12345""")
            conn.request.close()
        self.go([server_side], [client_side])

    def test_close_in_body(self):
        def client_side(client):
            exchange = client.exchange()
            self.check_exchange(exchange, {
                'version': b"1.1",
                'status': b"200",
                'phrase': b'OK',
            })

            @on(exchange)
            def error(err_msg):
                self.assertEqual(err_msg.__class__, thor.http.error.ConnectError)
                self.loop.stop()

            req_uri = b"http://%s:%i/close_in_body" % (test_host, test_port)
            exchange.request_start(b"GET", req_uri, [])
            exchange.request_done([])

        def server_side(conn):
            conn.request.send(b"""\
HTTP/1.1 200 OK
Content-Type: text/plain
Content-Length: 15
Connection: close

12345""")
            conn.request.close()
        self.go([server_side], [client_side])


    def test_conn_reuse(self):
        self.conn_checked = False
        def client_side(client):
            req_uri = b"http://%s:%i/conn_reuse" % (test_host, test_port)
            exchange1 = client.exchange()
            self.check_exchange(exchange1, {
                'version': b"1.1",
                'status': b"200",
                'phrase': b'OK',
                'body': b"12345"
            })

            @on(exchange1)
            def response_start(*args):
                self.conn_id = id(exchange1.tcp_conn)

            @on(exchange1)
            def response_done(trailers):
                exchange2 = client.exchange()
                self.check_exchange(exchange2, {
                    'version': b"1.1",
                    'status': b"404",
                    'phrase': b'Not Found',
                    'body': b"54321"
                })
                def start2():
                    exchange2.request_start(b"GET", req_uri, [])
                    exchange2.request_done([])
                self.loop.schedule(1, start2)

                @on(exchange2)
                def response_start(*args):
                    self.assertEqual(self.conn_id, id(exchange2.tcp_conn))
                    self.conn_checked = True

                @on(exchange2)
                def response_done(trailers):
                    self.loop.stop()

            exchange1.request_start(b"GET", req_uri, [])
            exchange1.request_done([])

        def server_side(conn):
            conn.request.sendall(b"""\
HTTP/1.1 200 OK
Content-Type: text/plain
Content-Length: 5

12345""")
            time.sleep(2)
            conn.request.sendall(b"""\
HTTP/1.1 404 Not Found
Content-Type: text/plain
Content-Length: 5
Connection: close

54321""")
            conn.request.close()
        self.go([server_side], [client_side])
        self.assertTrue(self.conn_checked)


    def test_conn_succeed_then_err(self):
        self.conn_checked = False
        def client_side(client):
            req_uri = b"http://%s:%i/succeed_then_err" % (test_host, test_port)
            exchange1 = client.exchange()
            self.check_exchange(exchange1, {
                'version': b"1.1",
                'status': b"200",
                'phrase': b'OK',
                'body': b"12345"
            })
            exchange2 = client.exchange()

            @on(exchange1)
            def response_start(*args):
                self.conn_id = id(exchange1.tcp_conn)

            @on(exchange1)
            def response_done(trailers):
                def start2():
                    exchange2.request_start(b"GET", req_uri, [])
                    exchange2.request_done([])
                self.loop.schedule(1, start2)

            @on(exchange2)
            def error(err_msg):
                self.conn_checked = True
                self.assertEqual(err_msg.__class__, thor.http.error.HttpVersionError)
                self.loop.stop()

            exchange1.request_start(b"GET", req_uri, [])
            exchange1.request_done([])

        def server_side(conn):
            conn.request.sendall(b"""\
HTTP/1.1 200 OK
Content-Type: text/plain
Content-Length: 5

12345""")
            time.sleep(2)
            conn.request.sendall(b"""\
HTTP/9.1 404 Not Found
Content-Type: text/plain
Content-Length: 5
Connection: close

54321""")
            conn.request.close()
        self.go([server_side], [client_side])
        self.assertTrue(self.conn_checked)


    def test_HEAD(self):
        def client_side(client):
            exchange = client.exchange()
            self.check_exchange(exchange, {
                'version': b"1.1",
                'status': b"200",
                'phrase': b'OK',
                'body': b""
            })
            @on(exchange)
            def response_done(trailers):
                self.loop.stop()

            req_uri = b"http://%s:%i/HEAD" % (test_host, test_port)
            exchange.request_start(b"HEAD", req_uri, [])
            exchange.request_done([])

        def server_side(conn):
            conn.request.send(b"""\
HTTP/1.1 200 OK
Content-Type: text/plain
Content-Length: 5
Connection: close

""")
            time.sleep(1)
            conn.request.close()
        self.go([server_side], [client_side])


    def test_req_retry(self):
        def client_side(client):
            exchange = client.exchange()
            self.check_exchange(exchange, {
                'version': b"1.1",
                'status': b"200",
                'phrase': b'OK',
                'body': b"12345"
            })
            @on(exchange)
            def response_done(trailers):
                self.loop.stop()

            req_uri = b"http://%s:%i/req_retry" % (test_host, test_port)
            exchange.request_start(b"OPTIONS", req_uri, [])
            exchange.request_done([])

        self.conn_num = 0
        def server_side(conn):
            self.conn_num += 1
            if self.conn_num > 1:
                conn.request.send(b"""\
HTTP/1.1 200 OK
Content-Type: text/plain
Content-Length: 5
Connection: close

12345""")
            conn.request.close()
        self.go([server_side], [client_side])


    def test_req_retry_fail(self):
        def client_side(client):
            exchange = client.exchange()
            self.check_exchange(exchange, {
                'version': b"1.1",
                'status': b"200",
                'phrase': b'OK',
                'body': b"12345"
            })

            @on(exchange)
            def error(err_msg):
                self.assertEqual(err_msg.__class__, thor.http.error.ConnectError)
                self.loop.stop()

            req_uri = b"http://%s:%i/req_retry_fail" % (test_host, test_port)
            exchange.request_start(b"OPTIONS", req_uri, [])
            exchange.request_done([])

        self.conn_num = 0
        def server_side(conn):
            self.conn_num += 1
            if self.conn_num > 3:
                conn.request.send(b"""\
HTTP/1.1 200 OK
Content-Type: text/plain
Content-Length: 5
Connection: close

12345""")
            conn.request.close()
        self.go([server_side], [client_side])


    def test_nobody(self):
        def client_side(client):
            exchange = client.exchange()
            self.check_exchange(exchange, {
                'version': b"1.1",
                'status': b"304",
                'phrase': b"Not Modified",
                'body': b''
            })

            @on(exchange)
            def response_done(trailers):
                self.loop.stop()

            req_uri = b"http://%s:%i" % (test_host, test_port)
            exchange.request_start(b"GET", req_uri, [])
            exchange.request_done([])

        def server_side(conn):
            conn.request.send(b"""\
HTTP/1.1 304 Not Modified
Content-Type: text/plain
Content-Length: 5
Connection: close

""")
            conn.request.close()
        self.go([server_side], [client_side])



    def test_nobody_body(self):
        def client_side(client):
            exchange = client.exchange()
            self.check_exchange(exchange, {
                'version': b"1.1",
                'status': b"304",
                'phrase': b"Not Modified",
                'body': b''
            })

            @on(exchange)
            def error(err_msg):
                self.assertEqual(err_msg.__class__, thor.http.error.ExtraDataError)
                self.loop.stop()

            req_uri = b"http://%s:%i" % (test_host, test_port)
            exchange.request_start(b"GET", req_uri, [])
            exchange.request_done([])

        def server_side(conn):
            conn.request.send(b"""\
HTTP/1.1 304 Not Modified
Content-Type: text/plain
Content-Length: 5
Connection: close

12345""")
            conn.request.close()
        self.go([server_side], [client_side])


    def test_extra_body(self):
        def client_side(client):
            exchange = client.exchange()
            self.check_exchange(exchange, {
                'version': b"1.1",
                'status': b"200",
                'phrase': b"OK",
                'body': b'12345'
            })

            @on(exchange)
            def error(err_msg):
                self.assertEqual(err_msg.__class__, thor.http.error.ExtraDataError)
                self.loop.stop()

            req_uri = b"http://%s:%i" % (test_host, test_port)
            exchange.request_start(b"GET", req_uri, [])
            exchange.request_done([])

        def server_side(conn):
            conn.request.send(b"""\
HTTP/1.1 200 OK
Content-Type: text/plain
Content-Length: 5
Connection: close

1234567890""")
            conn.request.close()
        self.go([server_side], [client_side])



# TODO:
#    def test_req_body(self):
#    def test_req_body_dont_retry(self):
#    def test_req_body_close_on_err(self):
#    def test_pipeline(self):
#    def test_malformed_hdr(self):
#    def test_unexpected_res(self):
#    def test_pause(self):
#    def test_options_star(self):
#    def test_idle_timeout(self):
#    def test_idle_timeout_reuse(self):
#    def test_alternate_tcp_client(self):

if __name__ == '__main__':
    unittest.main()




#!/usr/bin/env python

from __future__ import absolute_import
from distutils.core import setup
import thor

setup(
  name = 'thor',
  version = thor.__version__,
  description = 'Simple Event-Driven IO for Python',
  author = 'Mark Nottingham',
  author_email = 'mnot@mnot.net',
  url = 'http://github.com/mnot/thor/',
  download_url = \
    'http://github.com/mnot/thor/tarball/thor-%s' % thor.__version__,
  packages = ['thor', 'thor.http'],
  provides = ['thor'],
  long_description=open("README.rst").read(),
  license = "MIT",
  classifiers = [
    'Development Status :: 4 - Beta',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python',
    'Programming Language :: Python :: 3.5',
    'Operating System :: POSIX',
    'Topic :: Internet :: WWW/HTTP',
    'Topic :: Internet :: Proxy Servers',
    'Topic :: Internet :: WWW/HTTP :: HTTP Servers',
    'Topic :: Software Development :: Libraries :: Python Modules',
  ]
)


#!/usr/bin/env python

import sys
import unittest

from thor.http.common import header_names, header_dict, get_header

class TestHttpUtilsBytes(unittest.TestCase):
    hdrs = [
        (b'A', b'a1'),
        (b'B', b'b1'),
        (b'a', b'a2'),
        (b'C', b'c1'),
        (b'b', b'b2'),
        (b'A', b'a3, a4'),
        (b'D', b'"d1, d1"'),
    ]

    def test_header_names(self):
        hdrs_n = header_names(self.hdrs)
        self.assertEqual(hdrs_n, set([b'a', b'b', b'c', b'd']))

    def test_header_dict(self):
        hdrs_d = header_dict(self.hdrs)
        self.assertEqual(hdrs_d[b'a'], [b'a1', b'a2', b'a3', b'a4'])
        self.assertEqual(hdrs_d[b'b'], [b'b1', b'b2'])
        self.assertEqual(hdrs_d[b'c'], [b'c1'])

    def test_header_dict_omit(self):
        hdrs_d = header_dict(self.hdrs, b'b')
        self.assertEqual(hdrs_d[b'a'], [b'a1', b'a2', b'a3', b'a4'])
        self.assertTrue(b'b' not in list(hdrs_d))
        self.assertTrue(b'B' not in list(hdrs_d))
        self.assertEqual(hdrs_d[b'c'], [b'c1'])

    def test_get_header(self):
        self.assertEqual(get_header(self.hdrs, b'a'), [b'a1', b'a2', b'a3', b'a4'])
        self.assertEqual(get_header(self.hdrs, b'b'), [b'b1', b'b2'])
        self.assertEqual(get_header(self.hdrs, b'c'), [b'c1'])


class TestHttpUtilsStrings(unittest.TestCase):
    hdrs = [
        ('A', 'a1'),
        ('B', 'b1'),
        ('a', 'a2'),
        ('C', 'c1'),
        ('b', 'b2'),
        ('A', 'a3, a4'),
        ('D', '"d1, d1"'),
    ]

    def test_header_names(self):
        hdrs_n = header_names(self.hdrs)
        self.assertEqual(hdrs_n, set(['a', 'b', 'c', 'd']))

    def test_header_dict(self):
        hdrs_d = header_dict(self.hdrs)
        self.assertEqual(hdrs_d['a'], ['a1', 'a2', 'a3', 'a4'])
        self.assertEqual(hdrs_d['b'], ['b1', 'b2'])
        self.assertEqual(hdrs_d['c'], ['c1'])

    def test_header_dict_omit(self):
        hdrs_d = header_dict(self.hdrs, 'b')
        self.assertEqual(hdrs_d['a'], ['a1', 'a2', 'a3', 'a4'])
        self.assertTrue('b' not in list(hdrs_d))
        self.assertTrue('B' not in list(hdrs_d))
        self.assertEqual(hdrs_d['c'], ['c1'])

    def test_get_header(self):
        self.assertEqual(get_header(self.hdrs, 'a'), ['a1', 'a2', 'a3', 'a4'])
        self.assertEqual(get_header(self.hdrs, 'b'), ['b1', 'b2'])
        self.assertEqual(get_header(self.hdrs, 'c'), ['c1'])


if __name__ == '__main__':
    unittest.main()


#!/usr/bin/env python

import errno
import os
import socket
import sys
import tempfile
import time as systime
import unittest

from framework import make_fifo

import thor.loop


class IOStopper(thor.loop.EventSource):
    def __init__(self, testcase, loop):
        thor.loop.EventSource.__init__(self, loop)
        self.testcase = testcase
        self.r_fd, self.w_fd = make_fifo('tmp_fifo')
        self.on('fd_writable', self.write)
        self.register_fd(self.w_fd, 'fd_writable')

    def write(self):
        self.testcase.assertTrue(self._loop.running)
        self._loop.stop()
        os.close(self.r_fd)
        os.close(self.w_fd)
        os.unlink('tmp_fifo')


class TestLoop(unittest.TestCase):

    def setUp(self):
        self.loop = thor.loop.make()
        self.i = 0

    def increment_counter(self):
        self.i += 1

    def test_start_event(self):
        self.loop.on('start', self.increment_counter)
        self.loop.schedule(1, self.loop.stop)
        self.loop.run()
        self.assertEqual(self.i, 1)

    def test_stop_event(self):
        self.loop.on('stop', self.increment_counter)
        self.loop.schedule(1, self.loop.stop)
        self.loop.run()
        self.assertEqual(self.i, 1)

    def test_run(self):
        def check_running():
            self.assertTrue(self.loop.running)
        self.loop.schedule(0, check_running)
        self.loop.schedule(1, self.loop.stop)
        self.loop.run()

    def test_scheduled_stop(self):
        self.loop.schedule(1, self.loop.stop)
        self.loop.run()
        self.assertFalse(self.loop.running)

    def test_io_stop(self):
        r = IOStopper(self, self.loop)
        self.loop.run()
        self.assertFalse(self.loop.running)

    def test_run_stop_run(self):
        def check_running():
            self.assertTrue(self.loop.running)
        self.loop.schedule(0, check_running)
        self.loop.schedule(1, self.loop.stop)
        self.loop.run()
        self.assertFalse(self.loop.running)
        self.loop.schedule(0, check_running)
        self.loop.schedule(1, self.loop.stop)
        self.loop.run()

    def test_schedule(self):
        run_time = 3 # how long to run for
        def check_time(start_time):
            now = systime.time()
            self.assertTrue(
                now - run_time - start_time <= self.loop.precision,
                "now: %s run_time: %s start_time: %s precision: %s" % (
                    now, run_time, start_time, self.loop.precision
                )
            )
            self.loop.stop()
        self.loop.schedule(run_time, check_time, systime.time())
        self.loop.run()

    def test_schedule_delete(self):
        def not_good():
            assert Exception, "this event should not have happened."
        e = self.loop.schedule(2, not_good)
        self.loop.schedule(1, e.delete)
        self.loop.schedule(3, self.loop.stop)
        self.loop.run()

    def test_time(self):
        run_time = 2
        def check_time():
            self.assertTrue(
                abs(systime.time() - self.loop.time()) <= self.loop.precision
            )
            self.loop.stop()
        self.loop.schedule(run_time, check_time)
        self.loop.run()


class TestEventSource(unittest.TestCase):

    def setUp(self):
        self.loop = thor.loop.make()
        self.es = thor.loop.EventSource(self.loop)
        self.events_seen = []
        self.r_fd, self.w_fd = make_fifo('tmp_fifo')

    def tearDown(self):
        os.close(self.r_fd)
        os.close(self.w_fd)
        os.unlink('tmp_fifo')

    def test_EventSource_register(self):
        self.es.register_fd(self.r_fd)
        self.assertTrue(self.r_fd in list(self.loop._fd_targets))

    def test_EventSource_unregister(self):
        self.es.register_fd(self.r_fd)
        self.assertTrue(self.r_fd in list(self.loop._fd_targets))
        self.es.unregister_fd()
        self.assertFalse(self.r_fd in list(self.loop._fd_targets))

    def test_EventSource_event_del(self):
        self.es.register_fd(self.r_fd, 'fd_readable')
        self.es.on('fd_readable', self.readable_check)
        self.es.event_del('fd_readable')
        os.write(self.w_fd, b'foo')
        self.loop._run_fd_events()
        self.assertFalse('fd_readable' in self.events_seen)

    def test_EventSource_readable(self):
        self.es.register_fd(self.r_fd, 'fd_readable')
        self.es.on('fd_readable', self.readable_check)
        os.write(self.w_fd, b"foo")
        self.loop._run_fd_events()
        self.assertTrue('fd_readable' in self.events_seen)

    def test_EventSource_not_readable(self):
        self.es.register_fd(self.r_fd, 'fd_readable')
        self.es.on('fd_readable', self.readable_check)
        self.loop._run_fd_events()
        self.assertFalse('fd_readable' in self.events_seen)

    def readable_check(self, check=b"foo"):
        data = os.read(self.r_fd, 5)
        self.assertEqual(data, check)
        self.events_seen.append('fd_readable')

#    def test_EventSource_close(self):
#        self.es.register_fd(self.fd, 'close')
#        self.es.on('close', self.close_check)
#        self.fd.close()
#        self.loop._run_fd_events()
#        self.assertTrue('close' in self.events_seen)
#
#    def close_check(self):
#        self.events_seen.append('close')

if __name__ == '__main__':
    unittest.main()




#!/usr/bin/env python


"""
push-based asynchronous SSL/TLS-over-TCP

This is a generic library for building event-based / asynchronous
SSL/TLS servers and clients.
"""

import errno
import os
import socket
import ssl as sys_ssl

from thor.loop import LoopBase
from thor.tcp import TcpClient, TcpConnection

TcpConnection.block_errs.add(sys_ssl.SSL_ERROR_WANT_READ)
TcpConnection.block_errs.add(sys_ssl.SSL_ERROR_WANT_WRITE)
TcpConnection.close_errs.add(sys_ssl.SSL_ERROR_EOF)
TcpConnection.close_errs.add(sys_ssl.SSL_ERROR_SSL)

# TODO: TlsServer
# TODO: expose cipher info, peer info

class TlsClient(TcpClient):
    """
    An asynchronous SSL/TLS client.

    Emits:
      - connect (tcp_conn): upon connection
      - connect_error (err_type, err): if there's a problem before getting
        a connection. err_type is socket.error or socket.gaierror; err
        is the specific error encountered.

    To connect to a server:

    > c = TlsClient()
    > c.on('connect', conn_handler)
    > c.on('connect_error', error_handler)
    > c.connect(host, port)

    conn_handler will be called with the tcp_conn as the argument
    when the connection is made.
    """
    def __init__(self, loop: LoopBase=None) -> None:
        TcpClient.__init__(self, loop)
        try:
            self.tls_context = sys_ssl.SSLContext(sys_ssl.PROTOCOL_SSLv23)
        except AttributeError:
            self.tls_context = None

    def handshake(self) -> None:
        try:
            self.sock.do_handshake() # type: ignore
            self.once('fd_writable', self.handle_connect)
        except sys_ssl.SSLError as why:
            if isinstance(why, sys_ssl.SSLWantReadError):
#            if why == sys_ssl.SSL_ERROR_WANT_READ:
#                self.once('fd_readable', self.handshake)
                self.once('fd_writable', self.handshake) # Oh, Linux...
#            elif why == sys_ssl.SSL_ERROR_WANT_WRITE:
            elif isinstance(why, sys_ssl.SSLWantWriteError):
                self.once('fd_writable', self.handshake)
            else:
                self.handle_conn_error(sys_ssl.SSLError, why)
        except socket.error as why:
            self.handle_conn_error(socket.error, why)

    # TODO: refactor into tcp.py
    def connect(self, host: bytes, port: int, connect_timeout: float=None) -> None:
        """
        Connect to host:port (with an optional connect timeout)
        and emit 'connect' when connected, or 'connect_error' in
        the case of an error.
        """
        self.host = host
        self.port = port
        self.once('fd_writable', self.handshake)
        # FIXME: CAs
        if self.tls_context:
            self.sock = self.tls_context.wrap_socket( # type: ignore
                self.sock,
                do_handshake_on_connect=False,
                server_hostname=self.host
            )
        else: # server_hostname requires 2.7.9
            self.sock = sys_ssl.wrap_socket(
                self.sock,
                cert_reqs=sys_ssl.CERT_NONE,
                do_handshake_on_connect=False
            )
        # TODO: use socket.getaddrinfo(); needs to be non-blocking.
        try:
            err = self.sock.connect_ex((host, port))
        except socket.gaierror as why:
            self.handle_conn_error(socket.gaierror, why)
            return
        except socket.error as why:
            self.handle_conn_error(socket.error, why)
            return
        if err != errno.EINPROGRESS:
            self.handle_conn_error(socket.error, [err, os.strerror(err)])
            return
        if connect_timeout:
            self._timeout_ev = self._loop.schedule(
                connect_timeout,
                self.handle_conn_error,
                socket.error,
                [errno.ETIMEDOUT, os.strerror(errno.ETIMEDOUT)],
                True
            )


if __name__ == "__main__":
    import sys
    from thor import run
    test_host = sys.argv[1].encode('utf-8')

    def out(outbytes: bytes) -> None:
        sys.stdout.write(outbytes.decode('utf-8', 'replace'))

    def go(conn):
        conn.on('data', out)
        conn.write(b"GET / HTTP/1.1\r\nHost: %s\r\n\r\n" % test_host)
        conn.pause(False)
        print(conn.socket.cipher())

    c = TlsClient()
    c.on('connect', go)
    c.connect(test_host, 443)
    run()



#!/usr/bin/env python

"""
Thor shared HTTP infrastructure

This module contains utility functions and a base class
for the parsing portions of the HTTP client and server.
"""

from collections import defaultdict
from enum import Enum
from typing import Callable, Dict, List, Set, Tuple # pylint: disable=unused-import

from thor.http import error

linesep = b"\r\n"
RawHeaderListType = List[Tuple[bytes, bytes]]
OriginType = Tuple[bytes, bytes, int]

NEWLINE = ord("\n")
RETURN = ord("\r")

class Delimiters(Enum):
    CLOSE = 'close'
    COUNTED = 'counted'
    CHUNKED = 'chunked'
    NOBODY = 'nobody'

class States(Enum):
    WAITING = 1
    HEADERS_DONE = 2
    ERROR = 3
    QUIET = 4

idempotent_methods = [b'GET', b'HEAD', b'PUT', b'DELETE', b'OPTIONS', b'TRACE']
safe_methods = [b'GET', b'HEAD', b'OPTIONS', b'TRACE']
no_body_status = [b'204', b'304']
hop_by_hop_hdrs = [b'connection', b'keep-alive', b'proxy-authenticate',
                   b'proxy-authorization', b'te', b'trailers',
                   b'transfer-encoding', b'upgrade', b'proxy-connection']



def header_names(hdr_tuples: RawHeaderListType)-> Set[bytes]:
    """
    Given a list of header tuples, return the set of the header names seen.
    """
    return set([n.lower() for n, v in hdr_tuples])

def header_dict(hdr_tuples: RawHeaderListType, omit: List[bytes]=None) -> Dict[bytes, List[bytes]]:
    """
    Given a list of header tuples, return a dictionary keyed upon the
    lower-cased header names.

    If omit is defined, each header listed (by lower-cased name) will not be
    returned in the dictionary.
    """
    out = defaultdict(list)  # type: dict[bytes, list[bytes]]
    for (n, v) in hdr_tuples:
        n = n.lower()
        if n in (omit or []):
            continue
        if isinstance(v, bytes):
            splitter = b','
        else:
            splitter = ',' # type: ignore
        out[n].extend([i.strip() for i in v.split(splitter)])
    return out

def get_header(hdr_tuples: RawHeaderListType, name: bytes) -> List[bytes]:
    """
    Given a list of (name, value) header tuples and a header name (lowercase),
    return a list of all values for that header.

    This includes header lines with multiple values separated by a comma;
    such headers will be split into separate values. As a result, it is NOT
    safe to use this on headers whose values may include a comma (e.g.,
    Set-Cookie, or any value with a quoted string).
    """
    # TODO: support quoted strings
    if hdr_tuples and isinstance(hdr_tuples[0][1], bytes):
        splitter = b','
    else:
        splitter = ',' # type: ignore
    return [v.strip() for v in sum(
        [l.split(splitter) for l in [i[1] for i in hdr_tuples if i[0].lower() == name]], [])]


class HttpMessageHandler(object):
    """
    This is a base class for something that has to parse and/or serialise
    HTTP messages, request or response.

    For parsing, it expects you to override input_start, input_body and
    input_end, and call handle_input when you get bytes from the network.

    For serialising, it expects you to override output.

    Headers (and trailers) are represented as a list of (name, value) tuples,
    each being bytes.

    Body chunks are bytes.
    """

    careful = True # if False, don't fail on errors, but preserve them.
    default_state = None # type: States  # QUIET or WAITING

    def __init__(self) -> None:
        self.input_header_length = 0
        self.input_transfer_length = 0
        self._input_buffer = []  # type: list[bytes]
        self._input_state = self.default_state
        self._input_delimit = None  # type: Delimiters
        self._input_body_left = 0
        self._output_state = States.WAITING
        self._output_delimit = None  # type: Delimiters

    def __repr__(self) -> str:
        return "input %s output %s" % (self._input_state, self._output_state)

    # input-related methods

    def input_start(self, top_line: bytes, hdr_tuples: RawHeaderListType,
                    conn_tokens: List[bytes], transfer_codes: List[bytes],
                    content_length: int) -> Tuple[bool, bool]:
        """
        Take the top set of headers from the input stream, parse them
        and queue the request to be processed by the application.

        top_line is a bytes that contains the first line.

        hdr_tuples is a list of (bytes, bytes) tuples, one for each header line.

        conn_tokens is a list of bytes, one for each Connection token.

        transfer_codes is a list of bytes, one for each transfer coding.

        content_length is an integer representing the Content-Length header.

        Returns booleans (allows_body, is_final) to indicate whether the message allows a
        body, and whether it's the final message (respectively).

        Can raise ValueError to indicate that there's a problem and parsing
        cannot continue.
        """
        raise NotImplementedError

    def input_body(self, chunk: bytes) -> None:
        """
        Process a body chunk from the wire.

        Chunk is a bytes.
        """
        raise NotImplementedError

    def input_end(self, trailers: RawHeaderListType) -> None:
        """
        Indicate that the response body is complete. Optionally can contain
        trailers.

        Trailers are a list of (bytes, bytes) tuples, just like headers in _input_start.
        """
        raise NotImplementedError

    def input_error(self, err: error.HttpError) -> None:
        """
        Indicate an unrecoverable parsing problem with the input stream.

        err is an instance of thor.http.error.*.
        """
        raise NotImplementedError

    def handle_input(self, inbytes: bytes) -> None:
        """
        Given a bytes representing a chunk of input, figure out what state
        we're in and handle it, making the appropriate calls.
        """
        if self._input_buffer:
            self._input_buffer.append(inbytes)
            inbytes = b"".join(self._input_buffer)
            self._input_buffer = []
        if self._input_state == States.WAITING:  # waiting for headers or trailers
            headers, rest = self._split_headers(inbytes)
            if headers != None: # found one
                if self._parse_headers(headers):
                    try:
                        self.handle_input(rest)
                    except RuntimeError:
                        self.input_error(error.TooManyMsgsError())
                        # we can't recover from this, so we bail.
            else: # partial headers; store it and wait for more
                self._input_buffer.append(inbytes)
        elif self._input_state == States.QUIET:  # shouldn't be getting any data now.
            if inbytes.strip():
                self.input_error(error.ExtraDataError(inbytes.decode('utf-8', 'replace')))
        elif self._input_state == States.HEADERS_DONE:  # we found a complete header/trailer set
            try:
                body_handler = getattr(self, '_handle_%s' % self._input_delimit.value)
            except AttributeError:
                raise Exception("Unknown input delimiter %s" % \
                                 self._input_delimit)
            body_handler(inbytes)
        elif self._input_state == States.ERROR:  # something bad happened.
            pass # I'm silently ignoring input that I don't understand.
        else:
            raise Exception("Unknown state %s" % self._input_state)

    def _handle_nobody(self, inbytes: bytes) -> None:
        "Handle input that shouldn't have a body."
        self._input_state = self.default_state
        self.input_end([])
        self.handle_input(inbytes)

    def _handle_close(self, inbytes: bytes) -> None:
        "Handle input where the body is delimited by the connection closing."
        self.input_transfer_length += len(inbytes)
        self.input_body(inbytes)

    def _handle_chunked(self, inbytes: bytes) -> None:
        "Handle input where the body is delimited by chunked encoding."
        while inbytes:
            if self._input_body_left < 0: # new chunk
                inbytes = self._handle_chunk_new(inbytes)
            elif self._input_body_left > 0:
                # we're in the middle of reading a chunk
                inbytes = self._handle_chunk_body(inbytes)
            elif self._input_body_left == 0: # body is done
                self._handle_chunk_done(inbytes)
                break

    def _handle_chunk_new(self, inbytes: bytes) -> bytes:
        "Handle the start of a new body chunk."
        try:
            chunk_size, rest = inbytes.split(b"\r\n", 1)
        except ValueError:
            # don't have the whole chunk_size yet... wait a bit
            if len(inbytes) > 512:
                # OK, this is absurd...
                self.input_error(error.ChunkError(inbytes.decode('utf-8', 'replace')))
                # TODO: need testing around this; catching the right thing?
            else:
                self._input_buffer.append(inbytes)
            return b''
        # TODO: do we need to ignore blank lines?
        if b";" in chunk_size: # ignore chunk extensions
            chunk_size = chunk_size.split(b";", 1)[0]
        try:
            self._input_body_left = int(chunk_size, 16)
        except ValueError:
            self.input_error(error.ChunkError(chunk_size.decode('utf-8', 'replace')))
            return b''
        self.input_transfer_length += len(inbytes) - len(rest)
        return rest

    def _handle_chunk_body(self, inbytes: bytes) -> bytes:
        "Handle a continuing body chunk."
        got = len(inbytes)
        if self._input_body_left + 2 < got: # got more than the chunk
            this_chunk = self._input_body_left
            self.input_body(inbytes[:this_chunk])
            self.input_transfer_length += this_chunk + 2
            self._input_body_left = -1
            return inbytes[this_chunk + 2:] # +2 consumes the trailing CRLF
        elif self._input_body_left + 2 == got:
            # got the whole chunk exactly (including CRLF)
            self.input_body(inbytes[:-2])
            self.input_transfer_length += self._input_body_left + 2
            self._input_body_left = -1
        elif self._input_body_left == got: # corner case
            self._input_buffer.append(inbytes)
        else: # got partial chunk
            self.input_body(inbytes)
            self.input_transfer_length += got
            self._input_body_left -= got
        return b''

    def _handle_chunk_done(self, inbytes: bytes) -> None:
        "Handle a finished body chunk."
        if inbytes[:2] == b"\r\n": # no trailer
            self._input_state = self.default_state
            self.input_end([])
            if len(inbytes) > 2:
                self.handle_input(inbytes[2:]) # 2 consumes the CRLF
        else:
            trailer_block, rest = self._split_headers(inbytes) # trailers
            if trailer_block != None:
                self._input_state = self.default_state
                trailers = self._parse_fields(trailer_block.splitlines())
                if trailers is None: # found a problem
                    self._input_state = States.ERROR # TODO: need an explicit error
                    return
                else:
                    self.input_end(trailers)
                    self.handle_input(rest)
            else: # don't have full trailers yet
                self._input_buffer.append(inbytes)

    def _handle_counted(self, inbytes: bytes) -> None:
        "Handle input where the body is delimited by the Content-Length."
        if self._input_body_left <= len(inbytes): # got it all (and more?)
            self.input_transfer_length += self._input_body_left
            self.input_body(inbytes[:self._input_body_left])
            self.input_end([])
            self._input_state = self.default_state
            if inbytes[self._input_body_left:]:
                self.handle_input(inbytes[self._input_body_left:])
        else: # got some of it
            self.input_body(inbytes)
            self.input_transfer_length += len(inbytes)
            self._input_body_left -= len(inbytes)

    def _parse_fields(self, header_lines: List[bytes], gather_conn_info: bool=False):
        """
        Given a list of raw header lines (without the top line,
        and without the trailing CRLFCRLF), return its header tuples.
        """

        hdr_tuples = []        # type: RawHeaderListType
        conn_tokens = []       # type: list[bytes]
        transfer_codes = []    # type: list[bytes]
        content_length = None  # type: int

        for line in header_lines:
            if line[:1] in [b" ", b"\t"]: # Fold LWS
                if len(hdr_tuples):
                    hdr_tuples[-1] = (
                        hdr_tuples[-1][0],
                        b"%s %s" % (hdr_tuples[-1][1], line.lstrip())
                    )
                    continue
                else: # top header starts with whitespace
                    self.input_error(error.TopLineSpaceError(line.decode('utf-8', 'replace')))
                    if self.careful:
                        return
            try:
                fn, fv = line.split(b":", 1)
            except ValueError:
                continue # TODO: error on unparseable field?
            # TODO: a zero-length name isn't valid
            if fn[-1:] in [b" ", b"\t"]:
                self.input_error(error.HeaderSpaceError(fn.decode('utf-8', 'replace')))
                if self.careful:
                    return
            hdr_tuples.append((fn, fv))

            if gather_conn_info:
                f_name = fn.strip().lower()
                f_val = fv.strip()

                # parse connection-related headers
                if f_name == b"connection":
                    conn_tokens += [
                        v.strip().lower() for v in f_val.split(b',')
                    ]
                elif f_name == b"transfer-encoding": # TODO: parameters? no...
                    transfer_codes += [v.strip().lower() for \
                                       v in f_val.split(b',')]
                elif f_name == b"content-length":
                    if content_length != None:
                        try:
                            if int(f_val) == content_length:
                                # we have a duplicate, non-conflicting c-l.
                                continue
                        except ValueError:
                            pass
                        self.input_error(error.DuplicateCLError())
                        if self.careful:
                            return
                    try:
                        content_length = int(f_val)
                        assert content_length >= 0
                    except (ValueError, AssertionError):
                        self.input_error(error.MalformedCLError(f_val.decode('utf-8', 'replace')))
                        if self.careful:
                            return

        # yes, this is a horrible hack.
        if gather_conn_info:
            return hdr_tuples, conn_tokens, transfer_codes, content_length
        else:
            return hdr_tuples

    def _split_headers(self, inbytes: bytes) -> Tuple[bytes, bytes]:
        """
        Given a bytes, split out and return (headers, rest),
        consuming the whitespace between them.

        If there is not a complete header block, return None for headers.
        """

        pos = 0
        size = len(inbytes)
        while pos <= size:
            pos = inbytes.find(b"\n", pos)
            back = 0
            if pos == -1:
                return None, inbytes
            if pos > 0 and inbytes[pos - 1] == RETURN:
                back += 1
            pos += 1
            if pos < size:
                if inbytes[pos] == RETURN:
                    pos += 1
                    back += 1
                if pos < size and inbytes[pos] == NEWLINE:
                    return inbytes[:pos - back - 1], inbytes[pos + 1:]
        return None, inbytes

    def _parse_headers(self, inbytes: bytes) -> bool:
        """
        Given a bytes that we knows starts with a header block,
        parse the headers. Calls self.input_start to kick off processing.

        Returns True if no fatal problems are found.
        """
        self.input_header_length = len(inbytes)
        header_lines = inbytes.splitlines()

        # chop off the top line
        while True: # TODO: limit?
            try:
                top_line = header_lines.pop(0)
                if top_line.strip() != b"":
                    break
            except IndexError: # empty
                return True

        try:
            hdr_tuples, conn_tokens, transfer_codes, content_length \
                = self._parse_fields(header_lines, True)
        except TypeError: # returned None because there was an error
            return False # throw away the rest

        # ignore content-length if transfer-encoding is present
        if transfer_codes != [] and content_length != None:
            content_length = None

        try:
            allows_body, is_final = self.input_start(top_line, hdr_tuples, conn_tokens,
                                                     transfer_codes, content_length)
        except ValueError: # fatal parsing error of some kind; abort.
            return False # throw away the rest

        if not is_final:
            self._input_state = States.WAITING
        else:
            self._input_state = States.HEADERS_DONE
        if not allows_body:
            self._input_delimit = Delimiters.NOBODY
        elif len(transfer_codes) > 0:
            if transfer_codes[-1] == b'chunked':
                self._input_delimit = Delimiters.CHUNKED
                self._input_body_left = -1 # flag that we don't know
            else:
                self._input_delimit = Delimiters.CLOSE
        elif content_length != None:
            self._input_delimit = Delimiters.COUNTED
            self._input_body_left = content_length
        else:
            self._input_delimit = Delimiters.CLOSE
        return True

    ### output-related methods

    def output(self, out: bytes) -> None:
        """
        Write something to whatever we're talking to. Should be overridden.
        """
        raise NotImplementedError

    def output_start(self, top_line: bytes, hdr_tuples: RawHeaderListType, delimit: Delimiters):
        """
        Start outputting a HTTP message.
        """
        self._output_delimit = delimit
        out = [top_line]
        out.extend([b"%s: %s" % (k.strip(), v) for k, v in hdr_tuples])
        out.extend([b"", b""])
        self.output(linesep.join(out))
        self._output_state = States.HEADERS_DONE

    def output_body(self, chunk: bytes) -> None:
        """
        Output a part of a HTTP message. Takes bytes.
        """
        if not chunk or self._output_delimit is None:
            return
        if self._output_delimit == Delimiters.CHUNKED:
            chunk = b"%s\r\n%s\r\n" % (hex(len(chunk))[2:].encode('ascii'), chunk)
        self.output(chunk)
        # TODO: body counting
#        self._output_body_sent += len(chunk)
#        assert self._output_body_sent <= self._output_content_length, \
#            "Too many body bytes sent"

    def output_end(self, trailers: RawHeaderListType) -> None:
        """
        Finish outputting a HTTP message, including trailers if appropriate.
        """
        if self._output_delimit == Delimiters.NOBODY:
            pass # didn't have a body at all.
        elif self._output_delimit == Delimiters.CHUNKED:
            self.output(b"0\r\n%s\r\n" % b"\r\n".join([
                b"%s: %s" % (k.strip(), v) for k, v in trailers
            ]))
        elif self._output_delimit == Delimiters.COUNTED:
            pass # TODO: double-check the length
        elif self._output_delimit == Delimiters.CLOSE:
            self.tcp_conn.close() # type: ignore  pylint: disable=E1101
        elif self._output_delimit is None:
            pass # encountered an error before we found a delmiter
        else:
            raise AssertionError("Unknown request delimiter %s" % self._output_delimit)
        self._output_state = States.WAITING



#!/usr/bin/env python

import sys
import unittest

from thor.events import EventEmitter, on

class TestEventEmitter(unittest.TestCase):
    def setUp(self):
        class Thing(EventEmitter):
            def __init__(self):
                EventEmitter.__init__(self)
                self.foo_count = 0
                self.bar_count = 0
                self.rem1_count = 0
                self.rem2_count = 0
                self.on('foo', self.handle_foo)
                self.once('bar', self.handle_bar)
                self.on('baz', self.handle_baz)
                self.on('rem1', self.handle_rem1)
                self.on('rem1', self.handle_rem1a)
                self.on('rem2', self.handle_rem2)
                self.on('rem2', self.handle_rem2a)

            def handle_foo(self):
                self.foo_count += 1

            def handle_bar(self):
                self.bar_count += 1

            def handle_baz(self):
                raise Exception("Baz wasn't removed.")

            def handle_rem1(self):
                self.rem1_count += 1
                self.removeListeners()
                self.emit('foo')

            def handle_rem1a(self):
                self.rem1_count += 1

            def handle_rem2(self):
                self.rem2_count += 1
                self.removeListener('rem2', self.handle_rem2a)

            def handle_rem2a(self):
                self.rem2_count += 1

        self.t = Thing()

    def test_basic(self):
        self.assertEqual(self.t.foo_count, 0)
        self.t.emit('foo')
        self.assertEqual(self.t.foo_count, 1)
        self.t.emit('foo')
        self.assertEqual(self.t.foo_count, 2)

    def test_once(self):
        self.assertEqual(self.t.bar_count, 0)
        self.t.emit('bar')
        self.assertEqual(self.t.bar_count, 1)
        self.t.emit('bar')
        self.assertEqual(self.t.bar_count, 1)

    def test_removeListener(self):
        self.t.removeListener('foo', self.t.handle_foo)
        self.t.emit('foo')
        self.assertEqual(self.t.foo_count, 0)

    def test_removeListeners_named(self):
        self.t.removeListeners('baz')
        self.t.emit('baz')

    def test_removeListeners_named_multiple(self):
        self.t.removeListeners('baz', 'foo')
        self.t.emit('baz')
        self.t.emit('foo')
        self.assertEqual(self.t.foo_count, 0)

    def test_removeListeners_all(self):
        self.t.emit('foo')
        self.t.removeListeners()
        self.t.emit('foo')
        self.assertEqual(self.t.foo_count, 1)
        self.t.emit('baz')

    def test_sink(self):
        class TestSink(object):
            def __init__(self):
                self.bam_count = 0
            def bam(self):
                self.bam_count += 1
        s = TestSink()
        self.t.sink(s)
        self.assertEqual(s.bam_count, 0)
        self.t.emit('bam')
        self.assertEqual(s.bam_count, 1)
        self.assertEqual(self.t.foo_count, 0)
        self.t.emit('foo')
        self.assertEqual(self.t.foo_count, 1)

    def test_on_named(self):
        self.t.boom_count = 0
        @on(self.t, 'boom')
        def do():
            self.t.boom_count += 1
        self.assertEqual(self.t.boom_count, 0)
        self.t.emit('boom')
        self.assertEqual(self.t.boom_count, 1)

    def test_on_default(self):
        self.t.boom_count = 0
        @on(self.t)
        def boom():
            self.t.boom_count += 1
        self.assertEqual(self.t.boom_count, 0)
        self.t.emit('boom')
        self.assertEqual(self.t.boom_count, 1)


    def test_removeListeners_recursion(self):
        """
        All event listeners are called for a given
        event, even if one of the previous listeners
        calls removeListeners().
        """
        self.assertEqual(self.t.rem1_count, 0)
        self.t.emit('rem1')
        self.assertEqual(self.t.foo_count, 0)
        self.assertEqual(self.t.rem1_count, 2)

    def test_removeListener_recursion(self):
        """
        Removing a later listener specifically for
        a given event causes it not to be run.
        """
        self.assertEqual(self.t.rem2_count, 0)
        self.t.emit('rem2')
        self.assertEqual(self.t.rem2_count, 1)

if __name__ == '__main__':
    unittest.main()